/*
   Copyright 2000, 2001, 2002, 2003 Laurent Wacrenier

   This file is part of PLL lmtpd.

   PLL lmtpd is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   PLL lmtpd is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with PLL lmtpd; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

/* References :
 *
 * RFC 2104 - HMAC: Keyed-Hashing for Message Authentication
 * RFC 2595 - Using TLS with IMAP, POP3 and ACAP
 * RFC 2195 - IMAP/POP AUTHorize Extension for Simple Challenge/Response
 * RFC 2554 - SMTP Service Extension for Authentication
 *
 */
 

#include "config.h"

static char const rcsid[] UNUSED =
"$Id: auth.c,v 1.12 2005/08/03 11:04:39 lwa Exp $";


#include <syslog.h>
#include <strings.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include <ctype.h>
#include <errno.h>
#include "lmtp.h"
#include "rfc/rfc.h"
#include "libmd5/md5.h"

#define HMAC_IPAD 0x36
#define HMAC_OPAD 0x5C

/* see rfc 2104 */
/* md5(key ^ opad, md5(key ^ ipad, text)) */
static char *hmac_md5(char *key, char *text, char hexbuf[33]) {
  static const char hex[]="0123456789abcdef";
  md5_byte_t kipad[64];
  md5_byte_t kopad[64];
  unsigned char digest[16];
  md5_state_t pms;
  int i;
  memset(kipad, HMAC_IPAD, sizeof(kipad));
  memset(kopad, HMAC_OPAD, sizeof(kopad));
  for (i = 0; i<64 && key[i] ; i++) {
    kipad[i] ^= key[i];
    kopad[i] ^= key[i];
  }
  md5_init(&pms);
  md5_append(&pms, kipad, sizeof(kipad));
  md5_append(&pms, (md5_byte_t *)text, strlen(text));
  md5_finish(&pms, digest); 

  md5_init(&pms);
  md5_append(&pms, kopad, sizeof(kopad));
  md5_append(&pms, digest, sizeof(digest));
  md5_finish(&pms, digest);

  for(i = 0; i < 16; i++) {
    hexbuf[i+i] = hex[digest[i] >> 4];
    hexbuf[i+i+1] = hex[digest[i] & 0x0f];
  }
  hexbuf[i+i] = '\0';
  return hexbuf;
}

static char *auth_success(struct fstate *state, char *user, char *method) {
  state->sasl_method = NULL;
  if (state->sasl_user) 
    free(state->sasl_user);
  state->sasl_user = user;
  if (state->sasl_user == NULL)
    return MemoryError(state);
  state->sasl_need_auth = 0;
  state->sasl_method = method;
  syslog(LOG_INFO, ID_FMT "user '%s' authenticated with %s method", state->counter, user, method);
  return seterr(state, 235, "2.0.0", "Authentication successful.");
}

/* see rfc 2595 */
static char *auth_plain(struct fstate *state, char *authstring) {
  char *authenticate;
  char *password;
  char *real_password;
  char *decoded_authstring;
  size_t dlen;
  while(isspace((int)*authstring)) authstring++;
  if ((decoded_authstring = base64_alloc_decode(authstring, &dlen)) == NULL) {
    return seterr(state, 501, "5.6.5", "Invalid auth sequence.");
  }
  if ((authenticate = memchr(decoded_authstring, 0, dlen))  == NULL) {
    free(decoded_authstring);
    return seterr(state, 501, "5.5.2", "Invalid auth sequence.");
  }
  authenticate ++;
  dlen -= authenticate - decoded_authstring;
  if ((password = memchr(authenticate, 0, dlen)) == NULL) {
    free(decoded_authstring);
    return seterr(state, 501, "5.5.2", "Invalid auth sequence..");
  }
  password ++;
  dlen -= password - authenticate;
  real_password = slist_get_value(&state->sasl_users, authenticate);
  if (real_password == NULL ||
      strlen(real_password) != dlen ||
      strncmp(real_password, password, dlen) != 0) {
    char *err = seterr(state, 535, "5.3.0",
		       "Authentication failed for user '%s'", authenticate);
    free(decoded_authstring);
    return err;
  }
  return auth_success(state, strdup(authenticate), "PLAIN");
}

/* see RFC 2195 */
static char *auth_cram_md5(struct fstate *state) {
  char challenge[1024];
  char *challenge64;
  char buf[LINELEN+1];
  char *ebuf;
  char mdbuf[33];
  char *authstring;
  char *digest;
  char *real_password;
  size_t authstring_l;

  /* output the challenge */
  snprintf(challenge, sizeof(challenge), "<%ld.%lu@%s>",
	   random(), (unsigned long)time(NULL), state->myhostname);
  challenge[sizeof(challenge) - 1] = 0;
  if ((challenge64 = base64_alloc_encode(challenge, strlen(challenge)))
      == NULL)
    return MemoryError(state);
  printf("334 %s\r\n", challenge64);
  fflush(stdout);
  free(challenge64);

  /* get answer */
  timeout(state, state->timeout_cmd, 421);
  if (fgets(buf, sizeof(buf)-1, stdin) == NULL) {
    timeout(state, state->timeout_cmd, 0);
    return seterr(state, 421, "4.5.0", "End of input flow");
  }
   timeout(state, state->timeout_cmd, 0);
  if (exiting(state)) {
    return state->error;
  }
  ebuf = memchr(buf, '\n', LINELEN-1);
  if (ebuf == NULL) 
    return seterr(state, 500, "5.5.2", "command line too long");
  *ebuf -- = 0;
  while (ebuf >=buf && isspace((int)*ebuf))
    *ebuf--=0;

  /* get user and digest from answer */
  if ((authstring = base64_alloc_decode(buf, &authstring_l)) == NULL) {
    if (errno == ENOMEM)
      return MemoryError(state);
    return seterr(state, 535, "5.3.0", "Invalid sequence");
  }
  /* base64_alloc_decode() always add a final \0 */
  digest = strchr(authstring, ' ');
  if (digest == NULL) {
    free(authstring);
    return seterr(state, 535, "5.3.0", "Authentication failed");
  }
  *digest++ = 0;

  /* check with real password */
  if ( (real_password = slist_get_value(&state->sasl_users, authstring)) 
      == NULL ) {
    free(authstring);
    return seterr(state, 535, "5.3.0", "Authentication failed");
  }
  hmac_md5(real_password, challenge, mdbuf);
  if (strcmp(mdbuf, digest) != 0) {
    char *err = seterr(state, 535, "5.3.0", "Authentication failed for '%s'", authstring);
    free(authstring);
    return err;
  }

  /*  success */
  return auth_success(state, authstring, "CRAM-MD5");
}

/* see RFC 2554 */
char *auth(struct fstate *state, char *method, char *param) {
  if (state->sasl_noplain == 0 && strcasecmp("PLAIN", method) == 0) {
    return auth_plain(state, param);
  } else if (strcasecmp("CRAM-MD5", method) == 0) {
    return auth_cram_md5(state);
  } else {
    return seterr(state, 504, "5.3.3", "Unrecognized authentication type %s.", method);
  }
}
