/*
   Copyright 2000, 2001, 2002, 2003 Laurent Wacrenier

   This file is part of PLL lmtpd.

   PLL lmtpd is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   PLL lmtpd is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with PLL lmtpd; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include "config.h"

static char const rcsid[] UNUSED =
"$Id: deliver.c,v 1.39 2004/10/26 09:59:20 lwa Exp $";


#include <sys/types.h>
#include <sys/stat.h>
#include <sys/uio.h>
#include <unistd.h>
#include <errno.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <time.h>

#include "lmtp.h"


#if !HAVE_SETEUID
static int seteuid(uid_t euid) {
  if (euid == (uid_t)-1) {
    errno = EINVAL;
    return -1;
  }
#if HAVE_SETREUID
  return setreuid(-1, euid);
#elif HAVE_SETRESUID
  return setresuid(-1, euid, -1);
#else
#error should have seteuid, setreuid or setesuid
#endif
}
#endif

#if !HAVE_SETEGID
static int setegid(gid_t egid) {
  if (egid == (gid_t)-1) {
    errno = EINVAL;
    return -1;
  }
#if HAVE_SETREGID
  return setregid(-1, egid);
#elif HAVE_SETRESGID
  return setresgid(-1, egid, -1);
#else
#error should have setegid, setregid or setesgid
#endif
}
#endif

int mkdirp(char *path, mode_t mode) {
  char *p, *p1=path;
  while ((p=strchr(p1, '/'))!=NULL) {
    *p=0;
    mkdir(path, mode);
    *p='/';
    p1=p+1;
  }
  return mkdir(path, mode);
}

#define BUFLEN 8192

int setuidgid(struct fstate *state, struct rcpt *rcpt1) {
  uid_t uid;
  gid_t gid;
  char *user="";
  uid=rcpt1->uid;
  gid=rcpt1->gid;

  if (rcpt1->mbox)
    user=rcpt1->mbox;

  seteuid(getuid());

  if (setegid(gid)==-1) {
    seterr(state, 451, "4.3.0", "<%s>: setegid(%d): %s",
	   user, (int)gid, strerror(errno));
    return -1;
  }
  if (seteuid(uid)==-1) {
    seterr(state, 451, "4.3.0", "<%s>: seteuid(%d): %s",
	   user, (int)uid, strerror(errno));
    return -1;
  }
  return 0;
}

char *deliver(struct fstate *state, struct rcpt *rcpt1) {
  char *err;
  struct stat sb;

  if (state->fatal) {
    return state->error;
  }

  if (chdir(rcpt1->path)==-1) {
    if (rcpt1->shared) {
      return seterr(state, 451, "4.3.0", "<%s>: chdir: %s", rcpt1->user,
		    strerror(errno));
    }
    if (errno == ENOENT &&
	state->missing_user.uid != (uid_t)-1 &&
	state->missing_user.gid != (gid_t)-1 ) {
      if (setuidgid(state, &state->missing_user)==-1)
	return state->error;
      mkdirp(rcpt1->path, state->mode_hier);
      setuid(0);
      chown(rcpt1->path, rcpt1->uid, rcpt1->gid);
      if (setuidgid(state, rcpt1)==-1)
	return state->error;
      if (chdir(rcpt1->path)==-1) {
	return seterr(state, 451, "4.2.0", "<%s>: chdir: %s", rcpt1->user,
		      strerror(errno));
      }
    } else {
      return seterr(state, 451, "4.2.1",
		    "<%s>: unable to access mailbox: %s", rcpt1->user,
		    strerror(errno));
    }
  }

  if (!rcpt1->shared) {
    if (chdir("Maildir") == -1) {
      mkdir("Maildir", state->mode_mail);
      if (chdir("Maildir")==-1) {
	return seterr(state, 451, "4.3.0", "<%s>: chdir Maildir: %s",
		      rcpt1->user, strerror(errno));
      }
    }
  }

  if ((state->qrecipient=quote_addr821(rcpt1->user))==NULL) {
    return MemoryError(state);
  }

  state->rcpt=rcpt1;

  if (loopchk(rcpt1->user, state->hdr_delivered->next) < 0) {
    struct actionq act = {0,};
    logit(state, &act, 
	  ACTION_ERROR|ACTION_BOUNCE, "error (mail loop back to me)");
    logflush(state, &act);
    return seterr(state, 554, "5.4.6", "<%s>: mailbox looping",
		  state->qrecipient);
  } else {
    /* set Delivered-To header */
    int dtlen = sizeof("Delivered-To: ") + strlen(state->qrecipient);
    char *dt = realloc(state->hdr_delivered->header, dtlen);
    if (dt == NULL)
      return MemoryError(state);
    state->hdr_delivered->header = dt;
    strcpy(dt + sizeof("Delivered-To: ")-1, state->qrecipient);
    state->hdr_delivered->len = dtlen -1;
  }

  if (lstat("tmp", &sb)==-1) {
    if (errno==ENOENT) {
      char *md[]={ "new", "cur", "tmp" };
      int i;
      for (i=0; i<=2; i++)
	if (mkdir(md[i], state->mode_mail)==-1 && errno != EEXIST)
	  return seterr(state, 451, "4.3.0",
			"<%s>: unable to make '%s' in Maildir: %s",
			rcpt1->user, md[i],  strerror(errno));
    } else
      return seterr(state, 451, "4.3.0",
		    "<%s>: unable to access 'tmp' in Maildir: %s",
		    rcpt1->user, strerror(errno));
  } else {
    if ((sb.st_mode & (S_IWUSR|S_IXUSR|S_IFDIR)) != (S_IWUSR|S_IXUSR|S_IFDIR)
	|| sb.st_uid != rcpt1->uid )
      return seterr(state, 451, "4.3.0",
		    "<%s>: 'tmp' in Maildir: invalid permissions (%x,%x)",
		    rcpt1->user, (int)sb.st_mode&(S_IWUSR|S_IXUSR|S_IFDIR), 
		    S_IWUSR|S_IXUSR|S_IFDIR);
  }

  if (rcpt1->quota>0) {
    char buf[1000];
    snprintf(buf, sizeof(buf)-1, "%ldS", (long)rcpt1->quota);
    state->quotainfo = mdq_open(".", buf);

    if ( mdq_test(state->quotainfo, 0, 0) == -1) {
	int delay = quotawarn(state, rcpt1);
	int code = 452;
	fprintf(stderr, "delay=%d\n", delay);
	if (delay >= 0) {
	  char *a = acl_mailbox(&(state->quota_timefatal_acl), rcpt1->mbox);
	  int timefatal = state->quota_timefatal;
	  if (a)
	    timefatal = a2delay(a);
	  fprintf(stderr, "timefatal=%d\n", timefatal);
	  if (timefatal >= 0 &&
	      delay >= timefatal) 
	    code = 552;
	}
	quotadb_add(state, rcpt1->rcpt, code);
	return seterr(state, code, "4.2.2",
		      "<%s>: Mailbox full -- Quota exceeded",
		      state->qrecipient);
    } else {
      if (state->quota_mark) {
	if (unlink(state->quota_mark)!=-1)
	  quotadb_del(state, rcpt1->rcpt);
      }
    }
  } else {
    state->quotainfo = NULL;
    if (unlink("maildirsize")!=-1) {  /* just in case ... */
      if (state->quota_mark) {
	if (unlink(state->quota_mark)!=-1)
	  quotadb_del(state, rcpt1->rcpt);
      }
    }
  }

  if ((err=filter(state))==NULL)
    if ((err=exiting(state))==NULL)
      err=OkNoError(state);

  return err;
}
