/*
   Copyright 2000, 2001, 2002, 2003 Laurent Wacrenier

   This file is part of PLL lmtpd.

   PLL lmtpd is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   PLL lmtpd is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with PLL lmtpd; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include "config.h"

static char const rcsid[] UNUSED =
"$Id: lmtp.c,v 1.81 2005/04/26 09:42:56 lwa Exp $";



#include <sys/types.h>
#include <sys/socket.h>
#include <sys/stat.h>

#include <netdb.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <fcntl.h>


#include <locale.h>
#include <syslog.h>
#include <errno.h>

#if HAVE_LIBHOME
#include <hpwd.h>
#endif

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#if HAVE_SETPROCTITLE
#if HAVE_LIBUTIL_H
#include <libutil.h>
#endif
#endif

#include "rfc/rfc.h"
#include "lmtp.h"


#define LINELEN 1024

#define GET_SENDER      0
#define GET_DESTINATION 1

static int printerr(struct fstate *state, char *buf, char *err) {
  /*  char *err=state->error; */
  char *type;
  fputs(state->error, stdout);
  
  if (err == NULL)
    err = state->error;

#if WITH_MYSQLLOG
  mysqllog(state, buf);
#endif

  switch (*err) {
  case '4':
    type = "delay";
    break;
  case '5':
    if (state->logged_message) /* user bounce message */
      return 0;
    type= "reject";
    break;
  default:
    return 0;
  }
  if (state->fatal)
    type="fatal";
  syslog(LOG_INFO, ID_FMT "%s: %.*s; %.512s", 
	 state->counter, type, (int)strlen(err)-2,  err, buf);
  return state->fatal ? -1 : 0;
}

static int require_auth(struct fstate *state, char *buf) {
  if (state->sasl_need_auth && state->sasl_user == NULL) {
    seterr(state, 530, "5.1.7", "Authentication required");
    printerr(state, buf, NULL);
    return -1;
  }
  return 0;
}

static char *getuser_err(struct fstate *state, int sense) {
  if (sense==GET_SENDER) {
    seterr(state, 501, "5.1.7", "invalid sender mailbox address syntax");
  } else {
    seterr(state, 501, "5.1.3", "invalid destination mailbox address syntax");
  }
  return NULL;
}

static char *getuser(struct fstate *state, char *user, int sense) {
  char *unquoted;
  user += strspn(user, " ");

  if (*user++ != '<')
    return getuser_err(state, sense);

  if (sense == GET_SENDER && *user == '>') {
    char *unquoted=malloc(1);
    if (unquoted==NULL) {
      MemoryError(state);
      return NULL;
    }
    *unquoted=0;
    return unquoted;
  }
  unquoted=unquote_addr821(user, 
			   sense == GET_DESTINATION &&
			   state->no_domain_recipient ? 1 : 0,
			   &user);
  
  if (user==NULL) {
    MemoryError(state);
    return NULL;
  }

  if (unquoted==NULL || *user!='>') {
    free(unquoted);
    return getuser_err(state, sense);
  }

  if (sense==GET_SENDER) {
    user++;
    while(*user) {
      user += strspn(user, " ");
      if (!*user)
	break;
      if (strncasecmp(user, "SIZE=", 5)==0) {
	unsigned long size=strtoul(user+5, &user, 10);
	if (state->max_size > 0 && size > state->max_size) {
	  seterr(state, 552, "5.3.4", "Message size exceeds fixed limit");
	  free(unquoted);
	  return NULL;
	}
      }
      user += strcspn(user, " ");
    }
  }

  return unquoted;
}

static int go_to_spool(struct fstate *state) {
  if (setuidgid(state, &state->spool_user)==-1) {
    seterr(state, 421, "4.3.0", "cannot set spool privileges");
    return -1;
  }

  if (chdir(state->spool_user.path)==-1) {
    seterr(state, 421, "4.3.0", "cannot go to spool directory: %s",
	   strerror(errno));
    return -1;
  }
  return 0;
}

static void loop(struct fstate *state) {
  resetstate(state, STATE_CHAT);

#if HAVE_SETPROCTITLE
  setproctitle("%s", state->herhostname);
#endif

  while(!state->fatal) {
    char buf[LINELEN];
    char *ebuf;

    if (fflush(stdout)==EOF) {
      syslog(LOG_INFO, ID_FMT "output error: %m", state->counter);
      return;
    }

    timeout(state, state->timeout_cmd, 421);
    if (fgets(buf, sizeof(buf)-1, stdin) == NULL) {
      timeout(state, state->timeout_cmd, 0);
      return;
    }
    timeout(state, state->timeout_cmd, 0);
    ebuf = memchr(buf, '\n', LINELEN-1);
    if (ebuf == NULL) {
      seterr(state, 500, "5.5.2", "command line too long");
      printerr(state, buf, NULL);
      timeout(state, state->timeout_cmd, 421);
      /* read until the end of the line */
      while(fgets(buf, sizeof(buf)-1, stdin) &&
	    ! memchr(buf, '\n', LINELEN-1)) ; 
      timeout(state, state->timeout_cmd, 0);
      if (exiting(state)) 
	return;
      else
	continue; /* more luck next time ? */
    }
    do {
      *ebuf-- = 0;
    } while (ebuf >=buf && isspace((int)*ebuf));

    if (exiting(state)) {
      if (printerr(state, buf, NULL))
	return;
    }

    switch(*buf) {
    case 'd': case 'D':
      if (strcasecmp(buf, "data")==0) {
	if (state->nrcpts==0) {
	  seterr(state, 503, "5.5.1", "no recipients");
	  /*	  printerr(state, buf); */
	  fputs(state->error, stdout);
	  continue;
	}
	resetstate(state, STATE_DATA);
#if HAVE_SETPROCTITLE
	setproctitle("%s: MAIL FROM: <%s>, DATA",
		     state->herhostname, state->sender);
#endif
	if (prepare(state)!=NULL) {
	  printerr(state, buf, NULL);
	} else {
	  char *store_err;
	  int i;
	  int nbounce = 0;
	  int has_final_delivery = 0;
	  fputs("354 Start mail input; end with <CRLF>.<CRLF>\r\n", stdout);
	  fflush(stdout);
	  store_err=store(state);
	  
	  for (i=0; i<state->nrcpts; i++) {
	    if (store_err) {
	      printerr(state, buf, NULL);
	    } else {
	      struct rcpt *rcpt1 = state->rcpts+i;
	      if (rcpt1->bounce == NULL) {
		if (setuidgid(state, rcpt1)!=-1) {
		  gettimeofday(&state->date_deliver, NULL);
		  resetstate(state, STATE_DELIVER);
		  timeout(state, state->timeout_deliver, 450);
		  deliver(state, rcpt1);
		  timeout(state, state->timeout_deliver, 0);
		}
		if (state->can_bounce && state->error[0] == '5') {
		  rcpt1->bounce = strdup(state->error);
		}
		if (state->error[0] == '2' || state->error[0] == '5')
		  has_final_delivery = 1;
	      }
	      if (rcpt1->bounce) {
		nbounce ++;
		seterr(state, 200, "2.0.0", "will bounce: %s", rcpt1->bounce);
	      }
	      /* print returned message and log errors */
	      printerr(state, buf, rcpt1->bounce);
	      if (fflush(stdout)==EOF) {
		syslog(LOG_INFO, ID_FMT "output error: %m, "
		       "message may be sent more than once", state->counter);
		if (rcpt1->bounce) {
		  free(rcpt1->bounce);
		  rcpt1->bounce = 0;
		  nbounce --;
		}
		if (nbounce>0)
		  mybounce(state);
		return;
	      }
	    }
	  }
	  if (nbounce>0)
	    mybounce(state);
	  /* run antivirus on each mail */
	  if (has_final_delivery && state->av_checkall) { 
	    av_scan(state);
	  }
	  /* deliver viruses to virus keeper */
	  if (has_final_delivery && 
	      state->av.scanned && state->av.nvirus > 0 &&
	      state->av.viruskeeper.user) {
	    char *sender = state->sender;
	    resetstate(state, STATE_DELIVER);
	    state->sender = ""; /* avoid bounces on forward */
	    deliver(state, &state->av.viruskeeper);
	    state->sender = sender;
	  }
	  go_to_spool(state);
	}
	resetstate(state, STATE_CHAT);
#if HAVE_SETPROCTITLE
	setproctitle("%s", state->herhostname);
#endif
	continue;
      }
      break;
    case 'l': case 'L':
      if (strncasecmp(buf, "lhlo ", 5)==0) {
	char *lhlo;
	char *lhlo_c;

	free(state->lhlo);
	state->lhlo=NULL;
	lhlo   = buf+5;
	lhlo  += strspn(lhlo, " ");
	lhlo_c = lhlo;
	while(isalnum((int)*lhlo_c) || *lhlo_c=='-' || *lhlo_c=='.') lhlo_c++;
	
	if (lhlo_c == lhlo || (*lhlo_c && !isspace((int)*lhlo_c))) {
	  seterr(state, 501, "5.5.4", "invalid hostname");
	  printerr(state, buf, NULL);
	  continue;
	}

	*lhlo_c=0;
	state->lhlo=strdup(lhlo);

	printf("220-%s\r\n", state->myhostname);
	if (state->max_size>0) {
	  printf("220-SIZE %lu\r\n", (unsigned long)state->max_size);
	} else {
	  printf("220-SIZE\r\n");
	}
	if (state->sasl_users.nmemb > 0 || state->sasl_need_auth) {
	  if (state->sasl_noplain)
	    printf("220-AUTH CRAM-MD5\r\n");
	  else
	    printf("220-AUTH CRAM-MD5 PLAIN\r\n");
	}
	printf("220-8BITMIME\r\n"
	       "250-ENHANCEDSTATUSCODES\r\n"
	       "250 PIPELINING\r\n");
	continue;
      }
      break;
    case 'm': case 'M':
      if (require_auth(state, buf) == 0 && 
	  strncasecmp(buf, "mail from:", 10)==0) {
	if (state->state>=STATE_MAIL) {
	  seterr(state, 503, "5.5.1", "Nested MAIL command");
	  printerr(state, buf, NULL);
	  continue;
	}
	resetstate(state, STATE_MAIL);
	if ((state->sender=getuser(state, buf+10, GET_SENDER))!= NULL) {
	  printf("250 2.5.0 <%s> sender ok\r\n", state->sender);
#if HAVE_SETPROCTITLE
	  setproctitle("%s: MAIL FROM: <%s>", 
		       state->herhostname, state->sender);
#endif
	  if (state->mybounce && state->sender[0])
	    state->can_bounce = 1;
	  state->date=time(NULL);
	} else {
	  if (printerr(state, buf, NULL))
	    return;
	  resetstate(state, STATE_CHAT);
	}
	continue;
      }
      break;
    case 'r': case 'R':
      if (strncasecmp(buf, "rcpt to:", 8)==0) {
	char *user;
	if (!state->sender) {
	  seterr(state, 503, "5.5.1", "Need MAIL command");
	  printerr(state, buf, NULL);
	  continue;
	}
	user=getuser(state, buf+8, GET_DESTINATION);
	state->nrcpts_r++;
	if (user==NULL) {
	  if (printerr(state, buf, NULL))
	    return;
	} else if (newrcpt(state, user)!=NULL) {
	  free(user);
	  if (printerr(state, buf, NULL))
	    return;
	} else {
	  printf("250 2.1.5 <%s> recipient ok\r\n",
		 state->rcpts[state->nrcpts-1].user);
	}
	continue;
      } else if (strcasecmp(buf, "rset")==0) {
	fputs(seterr(state, 250, "2.0.0", "ok"), stdout);
#if HAVE_SETPROCTITLE
	setproctitle("%s", state->herhostname);
#endif
	resetstate(state, STATE_CHAT);
	continue;
      }
      break;
    case 'q': case 'Q':
      if (strcasecmp(buf, "quit")==0) {
	fputs(seterr(state, 221, "2.0.0", "bye"), stdout);
#if HAVE_SETPROCTITLE
	setproctitle("%s", state->herhostname);
#endif
	return;
      }
      break;
    case 'n': case 'N':
      if (strcasecmp(buf, "noop")==0) {
	fputs(seterr(state, 250, "2.0.0", "ok"), stdout);
	continue;
      }
      break;
    case 'v': case 'V':
      if (require_auth(state, buf) == 0 &&
	  strncasecmp(buf, "vrfy ", 5)==0) { /* VRFY is mandatory [RFC1123] */
	char *user=getuser(state, buf+5, GET_DESTINATION);
	if (user!=NULL)
	  do_vrfy(state, user);
	else
	  fputs(state->error, stdout);
	continue;
      }
      break;
    case 'e': case 'E':
      if (require_auth(state, buf) == 0 &&
	  strncasecmp(buf, "expn ", 5)==0) { /* EXPN is mandatory [RFC1123] */
	char *user=getuser(state, buf+5, GET_DESTINATION);
	if (user!=NULL)
	  do_expn(state, user);
	else
	  fputs(state->error, stdout);
	continue;
      }
      break;
    case 'a': case 'A':
      if (strncasecmp(buf, "auth ", 5)==0) {
	char *method = buf+5;
	char *param;
	while(isspace((int)*method)) method++;
	param = method;
	while(!isspace((int)*param)) param++;
	while(isspace((int)*param)) *param++ = 0;
	auth(state, method, param);
	printerr(state, buf, NULL);
	continue;
      }
      break;
    default:
      break;
    }
    seterr(state, 500, "5.5.2", "syntax error");
    printerr(state, buf, NULL);
  }
  return;
}

static int host_init(struct fstate *state) {
  char host[MAXHOSTNAMELEN+256];
  struct sockaddr sa;
  socklen_t sa_l=sizeof(sa);
  
  gethostname(host, sizeof(host)-1);
  state->myhostname=strdup(host);

  if (getpeername(0, &sa, &sa_l)==-1) {
    state->herhostname="stdin";
  } else {
    switch(sa.sa_family) {
    case AF_INET: {
      struct sockaddr_in *sa_in = (struct sockaddr_in *)&sa;
      struct hostent *hs=NULL;

      hs=gethostbyaddr((char *)&(sa_in->sin_addr),
		       sizeof(sa_in->sin_addr), AF_INET);
      if (hs) {
	snprintf(host, sizeof(host) -1, "%s [%s]", 
		 hs->h_name,
		 inet_ntoa(sa_in->sin_addr));
      } else {
	snprintf(host, sizeof(host) -1, "[%s]",
		 inet_ntoa(sa_in->sin_addr));
      }
      state->herhostname=strdup(host);
    }
      break;
    default:
      state->herhostname="localhost";
      break;
    }
  }
  syslog(LOG_INFO, ID_FMT "connect from %s",
	 state->counter, state->herhostname);
  return 0;
}

static char *progname(char *argv0) {
  char *end = strrchr(argv0, '/');
  if (end)
    return end+1;
  else
    return argv0;
}

int main(int argc, char **argv) {
  struct fstate state = {0,};
  char *conffile=CONFFILE;
  int ch;
  extern int optopt;
  extern char *optarg;
  extern int opterr;
  int logopt = LOG_PID;
  char *myname;
  int close_stderr = 1;

  opterr = 0;

  myname = progname(argv[0]);

  openlog(myname, logopt, LOG_MAIL);

  while ((ch = getopt(argc, argv, "VC:en:")) != -1) {
    switch (ch) {
    case 'C':
      conffile=optarg;
      break;
    case 'e':
#ifdef LOG_PERROR
      logopt |= LOG_PERROR;
      openlog(myname, logopt, LOG_MAIL);
#else 
      fprintf(stderr, "log to stderr unsupported on this plateform\n");
#endif
      close_stderr = 0;
      break;
    case 'n':
      myname = optarg;
      openlog(myname, logopt, LOG_MAIL);
      break;
    case 'V':
      version();
      exit(0);
    default:
      syslog(LOG_ERR, ID_FMT "illegal option '-%c'", 0UL, optopt);
      break;
    }
  }

  setsignals();

  resetstate(&state, STATE_INIT);

  if (readconf(conffile, &state)<0
      || mybounce_init(&state) < 0
#if WITH_PERL
      || fperl_init(&state) < 0
#endif
      ) {
    syslog(LOG_ERR, "configuration error, exiting");
    printf("421 4.3.5 configuration error\r\n");
    fflush(stdout);
    exit(0);
  }

  state.pid=(long)getpid();
  state.date_start=time(NULL);

  if (state.syslog_facility && state.syslog_facility != LOG_MAIL)
    openlog(myname, logopt, state.syslog_facility);

  host_init(&state);
  time_init(&state);
  if (rfc2047_init() != 0)
    exit(1);
  if (re_init() != 0)
    exit(1);


  av_init(&state);
  spam_assassin_init(&state);
  if (close_stderr) {
    int fd = open("/dev/null", O_WRONLY);
    if (fd == -1)
      syslog(LOG_ERR, ID_FMT "unable to open /dev/null: %s", 0UL, 
	     strerror(errno));
    else
      dup2(fd, 2);
  }


#if HAVE_SRANDOMDEV
  srandomdev();
#else
  srandom(state.date_start + state.pid);
#endif

  state.myname = myname;
#if HAVE_LIBHOME
  setpwtag(myname);
  setpassent(1);
  getpwnam(""); /* open the configuration file with root privileges */
#endif


  if (go_to_spool(&state) == -1) {
    fputs(state.error, stdout);
    exit(1);
  }

  quotadb_open(&state);

  printf("220 %s LMTP (%lu)\r\n", state.myhostname, (unsigned long)state.pid);
  loop(&state) ;
  closestate(&state); /* close files */
  exit(EXIT_SUCCESS);
}

