/*
   Copyright 2000, 2001, 2002, 2003 Laurent Wacrenier

   This file is part of PLL lmtpd.

   PLL lmtpd is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   PLL lmtpd is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with PLL lmtpd; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include "config.h"

static char const rcsid[] UNUSED =
"$Id: lmtpconf.c,v 1.74 2005/08/03 11:04:46 lwa Exp $";



#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>

#if HAVE_PATHS_H
#include <paths.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <syslog.h>
#include <ctype.h>
#include <errno.h>

#include "lmtp.h"

#define CONFLINEMAX 1024

#define FLAGS_STRING     0x01
#define FLAGS_INT        0x02
#define FLAGS_UID_T      0x03
#define FLAGS_GID_T      0x04
#define FLAGS_MODE_T     0x05
#define FLAGS_ULONG      0x06
#define FLAGS_TIME       0x07
#define FLAGS_SYSLOG     0x08
#define FLAGS_SLIST      0x09
#define FLAGS_SLIST_P    0x0A
#define FLAGS_CHAR       0x0B
#if HAVE_GETADDRINFO
#define FLAGS_UDP        0x0C
#endif
#define FLAGS_BOOL       0x0D
#define FLAGS_ARGS       0x0E
#define FLAGS_ARGSF      0x0F
#define FLAGS_RE         0x10
#define FLAGS_RELIST     0x11

#define FLAGS_INCLUDE    0x1F

#define FLAGS_MASK       0x00FF

#define FLAGS_OBSOLETE   0x0100

struct facility_list {
  char   *name;
  int     facility;
};

struct conf {
  const char *name;
  void *val;
  int type;
  char *def;
};


struct conf_state {
  int level;
  struct fstate *state;
  struct conf *spec;
};

/* from Postfix */
static struct facility_list facility_list[] = {
#ifdef LOG_AUTH
  { "auth", LOG_AUTH, },
#endif
#ifdef LOG_AUTHPRIV
  { "authpriv", LOG_AUTHPRIV, },
#endif
#ifdef LOG_CRON
  { "cron", LOG_CRON, },
#endif
#ifdef LOG_DAEMON
  { "daemon", LOG_DAEMON, },
#endif
#ifdef LOG_FTP
  { "ftp", LOG_FTP, },
#endif
#ifdef LOG_KERN
  { "kern", LOG_KERN, },
#endif
#ifdef LOG_LPR
  { "lpr", LOG_LPR, },
#endif
#ifdef LOG_MAIL
  { "mail", LOG_MAIL, },
#endif
#ifdef LOG_NEWS
  { "news", LOG_NEWS, },
#endif
#ifdef LOG_SECURITY
  { "security", LOG_SECURITY, },
#endif
#ifdef LOG_SYSLOG
  { "syslog", LOG_SYSLOG, },
#endif
#ifdef LOG_USER
  { "user", LOG_USER, },
#endif
#ifdef LOG_UUCP
  { "uucp", LOG_UUCP, },
#endif
#ifdef LOG_LOCAL0
  { "local0", LOG_LOCAL0, },
#endif
#ifdef LOG_LOCAL1
  { "local1", LOG_LOCAL1, },
#endif
#ifdef LOG_LOCAL2
  { "local2", LOG_LOCAL2, },
#endif
#ifdef LOG_LOCAL3
  { "local3", LOG_LOCAL3, },
#endif
#ifdef LOG_LOCAL4
  { "local4", LOG_LOCAL4, },
#endif
#ifdef LOG_LOCAL5
  { "local5", LOG_LOCAL5, },
#endif
#ifdef LOG_LOCAL6
  { "local6", LOG_LOCAL6, },
#endif
#ifdef LOG_LOCAL7
  { "local7", LOG_LOCAL7, },
#endif
  { NULL, 0 },
};

static int readconf1(char *file, struct fstate *state,
		     struct conf *spec);

static struct regexp_list *compile_relist(char *line, int flags) {
  struct regexp_list *rl;
  int ret;
  char *xline=line;
  while(*xline && (!isblank(*xline) || (xline >line && xline[-1]=='\\')))
    xline++;
  if (*xline) {
    *xline++=0;
    while(isblank(*xline))
      xline++;
  }

  rl=(struct regexp_list *)malloc(sizeof(struct regexp_list));
  if (rl==NULL) {
    syslog(LOG_WARNING, "unable to allocate memory in regexp list: %s", strerror(errno));
    return NULL;
  }

  ret=regcomp(&rl->preg, line, flags);

  if (ret) {
    char errbuf[CONFLINEMAX];
    regerror(ret, &rl->preg, errbuf, CONFLINEMAX);
    syslog(LOG_WARNING, "invalid regexp '%s': %s", line, errbuf);
    free(rl);
    return NULL;
  }

  rl->next=NULL;
  if ((rl->action=strdup(xline))==NULL) {
    free(rl);
    syslog(LOG_WARNING, "unable to allocate memory in regexp list: %s", strerror(errno));
    return NULL;
  }

  return rl;
}


static int setval(void *val, int type, char *args) {
  switch(type) {
  case FLAGS_INT:
    *(int *)val=args ? (int)strtol(args, NULL, 0) : 0;
    return 0;

  case FLAGS_ARGS: {      /* liste de mots */
    struct arg *vala = val;
    if (vala->argv) {
      if (vala->argv[0])
	free(vala->argv[0]);
      free(vala->argv);
    }
    vala->argc = 0;
    /* 2 entres pour commencer */
    vala->argv = malloc(sizeof(char**)*2);
    vala->argv[0] = NULL;
    if (args && args[0]) {
      if ((args = strdup(args))==NULL)
	return -1;
      
      vala->argv[0] = args;
      vala->argc = 1;
      while(*args) {
	if (isspace((int)*args)) {
	  while(isspace((int)*args)) *args++=0;
	  if (*args==0) break;
	  vala->argv[vala->argc++]=args;
	  vala->argv=(char**)realloc(vala->argv,
				     sizeof(char**)*(vala->argc+1));
	}
	args++;
      }
      vala->argv[vala->argc]=NULL;
    }
  }
    return 0;

  case FLAGS_CHAR:
    *(char *)val=args ? *args : 0;
    return 0;
  case FLAGS_TIME:
    *(int *)val=args ? (int)a2delay(args) : 0;
    return 0;
  case FLAGS_UID_T:
    *(uid_t *)val=args ? (uid_t)strtol(args, NULL, 0) : 0;
    return 0;
  case FLAGS_GID_T:
    *(gid_t *)val=args ? (gid_t)strtol(args, NULL, 0) : 0;
    return 0;
  case FLAGS_MODE_T:
    *(mode_t *)val=args ? (mode_t)strtol(args, NULL, 0) : 0;
    return 0;
  case FLAGS_ULONG:
    *(size_t *)val=args ? (size_t)strtoul(args, NULL, 0) : 0;
    return 0;
  case FLAGS_STRING:
    if (*(char **)val) free(*(char **)val);
    *(char **)val=args ? strdup(args) : NULL;
    return 0;
  case FLAGS_RE: {
    regex_t *preg = *(regex_t **)val;
    if (preg) { 
      regfree(preg);
      free(preg);
      preg = NULL;
    }
    if (args) {
      int ret;
      preg = malloc(sizeof(regex_t));
      ret = regcomp(preg, args, REG_EXTENDED);
      if (ret) {
	char errbuf[100];
	regerror(ret, preg, errbuf, sizeof(errbuf));
	syslog(LOG_WARNING, "invalid regexp %s", errbuf);
	return -1;
      }
    }
    *(regex_t **)val = preg;
    return 0;
  }

    
  case FLAGS_SYSLOG:
    if (args) {
      struct facility_list *fnp;
      for (fnp = facility_list; fnp->name; ++fnp) {
	if (!strcmp(fnp->name, args)) {
	  *(int *)val = fnp->facility;
	  return 0;
	}
      }
    } else {
	*(int *)val = 0;
    }
    return 0;
  case FLAGS_SLIST:
    if (args) {
      struct slist *m = (struct slist *)val;
      char *key = args;
      char *value = strpbrk(args, " \t:");
      if (value) {
	do {
	  *value ++ = 0;
	} while(*value == ' ' || *value == '\t');
      } else
	value = "";
      slist_put(m, key, value);
    } else {
      slist_clean((struct slist *)val);
    }
    return 0;
  case FLAGS_SLIST_P:
    if (args) {
      struct slist *m = *(struct slist **)val;
      char *key = args;
      char *value = strpbrk(args, " \t:");
      if (value) {
	do {
          *value ++ = 0;
        } while(*value == ' ' || *value == '\t');
      } else
	value = "";
      m = slist_put(m, key, value);
      if (m==NULL)
	return -1;
      *(struct slist **)val = m;
    } else {
      struct slist *m = *(struct slist **)val;
      slist_clean(m);
      *(struct slist **)val = NULL;
    }
    return 0;

  case FLAGS_RELIST:
    if (args) {
      struct regexp_list *m = *(struct regexp_list **)val;
      struct regexp_list *re = compile_relist(args, REG_EXTENDED|REG_NOSUB);
      if (re == NULL)
	return -1;
      if (m) {
	while(m->next)
	  m = m->next;
	m->next = re;
      } else {
	*(struct regexp_list **)val = re;
      }
    } else {
      struct regexp_list *m = *(struct regexp_list **)val;
      while(m) {
	struct regexp_list *next = m->next;
	regfree(&m->preg);
	free(m);
	m = next;
      }
    }

    return 0;

#if FLAGS_UDP
  case FLAGS_UDP:
    if (*(struct sockaddr **)val) free(*(struct sockaddr **)val);
    *(struct sockaddr **)val = NULL;
    if (args) {
      char *port = strchr(args, ':');
      struct addrinfo *res;
      struct addrinfo hints;
      int error;
      if (port)
	*port ++ = 0;
      memset(&hints, 0, sizeof(hints));
      hints.ai_family = AF_INET;
      hints.ai_socktype = SOCK_DGRAM;
      error = getaddrinfo(args, port, &hints, &res);
      if (error) 
	return -1;
      if (res) {
	struct sockaddr *sa = malloc(res->ai_addrlen);
	memcpy(sa, res->ai_addr, res->ai_addrlen);
	*(struct sockaddr **)val = sa;
      }
      freeaddrinfo(res);
    }
    return 0;
#endif
  case FLAGS_BOOL:
    if (args && (*args == 'y' || *args == 'Y' ||
		 *args == 't' || *args == 'T' || *args == '1'))
      *(int *)val = 1;
    else
      *(int *)val = 0;
    return 0;
  case FLAGS_INCLUDE: 
    if (args) {
      struct conf_state *cstate = val;
      char *filename = args;
      int ret = -1;
      cstate->level++;
      if (*filename) {
	if (cstate->level > MAX_INCLUDE_LEVEL) {
	  syslog(LOG_WARNING, "include too deep while including %s", filename);
	} else
	  ret = readconf1(filename, cstate->state, cstate->spec);
      }
      cstate->level--;
      return ret;
    } else
      return 0;
  }
  return -1;
}

static struct conf_state *resetconf(struct fstate *state) {
  struct conf_state *cstate = malloc(sizeof (struct conf_state));
  struct conf spec [] =  {
    { "syslog.acl",       &(state->syslog_acl),       FLAGS_SLIST , NULL},
#if FLAGS_UDP
    { "syslog.src",       &(state->syslog_src),       FLAGS_UDP,    NULL},
#endif
    { "rbl.server",       &(state->rbl_server),       FLAGS_SLIST , NULL},
    { "rhsbl.server",     &(state->rhsbl_server),     FLAGS_SLIST,  NULL},
    { "res.retry",        &(state->res_retry),        FLAGS_INT,    "0" },
    { "res.retrans",      &(state->res_retrans),      FLAGS_TIME,   "1" },
    { "rbl.res.retry",    &(state->res_retry),        
      FLAGS_INT | FLAGS_OBSOLETE,    "res.retry" },
    { "rbl.res.retrans",  &(state->res_retrans),      
      FLAGS_TIME | FLAGS_OBSOLETE,  "res.retrans" },

    { "reply.id",         &(state->reply_id),         FLAGS_BOOL,   NULL},
    { "filtername",       &(state->filtername),       FLAGS_ARGS,   NULL },
    { "log.name",         &(state->logname),          FLAGS_STRING, NULL },
    { "vacation.cache",   &(state->vacationcache),    FLAGS_STRING, NULL },
    { "vacation.cache.create",   &(state->vacationcache_create),
      FLAGS_BOOL, NULL },
    { "received.max",     &(state->max_Received),     FLAGS_INT,    "25" },
    { "received.front",   &(state->received_front),   FLAGS_RE,      NULL },
    { "timeout.cmd",      &(state->timeout_cmd),      FLAGS_TIME,    NULL },
    { "timeout.dataline", &(state->timeout_dataline), FLAGS_TIME,    NULL },
    { "timeout.deliver",  &(state->timeout_deliver),  FLAGS_TIME,    NULL },
    { "missing.path.uid", &(state->missing_user.uid), FLAGS_UID_T,  "65534" },
    { "missing.path.gid", &(state->missing_user.gid), FLAGS_UID_T,  "65534" },
    { "share.map",        &(state->share_map),        FLAGS_STRING, NULL },
    { "share.domain",     &(state->share_domain),     FLAGS_STRING, NULL },
    { "mode.hier",        &(state->mode_hier),        FLAGS_MODE_T, "0755" },
    { "mode.mail",        &(state->mode_mail),        FLAGS_MODE_T, "0751" },
    { "size.max",         &(state->max_size),         FLAGS_ULONG,  NULL },
    { "spooldir",         &(state->spool_user.path),  FLAGS_STRING, 
      "/var/spool/lmtp" },
    { "uid",              &(state->spool_user.uid),   FLAGS_UID_T, "65534" },
    { "gid",              &(state->spool_user.gid),   FLAGS_GID_T, "65534" },
    { "log.maxsize",      &(state->log_maxsize),      FLAGS_INT,    NULL },
    { "log.mintime",      &(state->log_mintime),      FLAGS_TIME,   NULL },
    { "log.msg",          &(state->log_msg),          FLAGS_STRING, NULL },
    { "syslog.facility",  &(state->syslog_facility),  FLAGS_SYSLOG, NULL},
    { "timezone",         &(state->timezone),         FLAGS_STRING, NULL},
    { "module.path",      &(state->module_path),      FLAGS_STRING, NULL},
    { "bodycheck.maxlen", &(state->bodycheck_maxlen), FLAGS_INT   , "0" },
    { "bounce.message",   &(state->bounce_message),   FLAGS_STRING,
      "Permission denied" },
    { "recipient.delimiter", &(state->recipient_delimiter), FLAGS_CHAR, "+"},
    { "domain.local",      &(state->localdomain),      FLAGS_STRING, NULL},
    { "no.domain.recipient", &(state->no_domain_recipient), FLAGS_BOOL, NULL},
    { "forwards.max",      &(state->forwards_max),     FLAGS_INT, "-1"},

    { "quota.warnmsg",    &(state->quota_warnmsg),    FLAGS_STRING, NULL },
    { "quota.mark",       &(state->quota_mark),       FLAGS_STRING, NULL },
    { "quota.timefatal",  &(state->quota_timefatal),  FLAGS_TIME, "-1" },
    { "quota.timefatal.acl", &(state->quota_timefatal_acl),
      FLAGS_SLIST,  NULL},
    { "quota.maxchecks",  &(state->quota_maxchecks),  FLAGS_TIME, "-1" },
    { "quota.link.dir",   &(state->quota_db_linkdir),  FLAGS_STRING, NULL},
    { "quota.cache.timeout", &(state->quota_db_timeout), FLAGS_TIME, "-1" },
    { "quota.cache.method",  &(state->quota_db_method), FLAGS_STRING, NULL},

    { "sasl.user",         &(state->sasl_users),      FLAGS_SLIST, NULL},
    { "sasl.need.auth",    &(state->sasl_need_auth),  FLAGS_BOOL, NULL},
    { "sasl.noplain",      &(state->sasl_noplain),    FLAGS_BOOL, NULL},
    { "mybounce",          &(state->mybounce),        FLAGS_BOOL, NULL},
    { "mybounce.msg",      &(state->mybounce_msg),    FLAGS_ARGS, NULL},
    { "mybounce.path",     &(state->mybounce_path),   FLAGS_STRING, NULL},
    { "sendmail.path",     &(state->sendmail_path),   FLAGS_STRING,
      _PATH_SENDMAIL },
    { "file.charset.re",   &(state->file_charset_re), FLAGS_RELIST, NULL},
#if WITH_DB
    { "quota.db3.file",      &(state->quota_db_dbfile), 
      FLAGS_STRING | FLAGS_OBSOLETE,  "quota.db.file" },
    { "quota.db3.env",       &(state->quota_db_dbenv),
      FLAGS_STRING | FLAGS_OBSOLETE,  "quota.db.env" },

    { "quota.db.file",      &(state->quota_db_dbfile),  FLAGS_STRING,
      "quota.db" },
    { "quota.db.env",       &(state->quota_db_dbenv),   FLAGS_STRING, NULL, },
#endif
    { "alias.file",         &(state->alias_file),       FLAGS_STRING, NULL, },
    { "alias.format",       &(state->alias_format),     FLAGS_RE, NULL, },
    { "alias.sub.addresses", &(state->alias_sub_addresses), FLAGS_INT, "-1", },
#if WITH_MYSQLLOG
    { "quota.mysql.host",    &(state->quota_db_mysql_host),
      FLAGS_STRING, NULL},
    { "quota.mysql.user",    &(state->quota_db_mysql_user), 
      FLAGS_STRING, NULL},
    { "quota.mysql.password", &(state->quota_db_mysql_password),
      FLAGS_STRING, NULL},
    { "quota.mysql.database", &(state->quota_db_mysql_database),
      FLAGS_STRING, NULL},
    { "quota.mysql.table",    &(state->quota_db_mysql_table),
      FLAGS_STRING, NULL},
    { "quota.mysql.col.user", &(state->quota_db_mysql_col_user),
      FLAGS_STRING, NULL},
    { "quota.mysql.col.code", &(state->quota_db_mysql_col_code),
      FLAGS_STRING, NULL},
    { "quota.mysql.col.time", &(state->quota_db_mysql_col_time),
      FLAGS_STRING, NULL},
    { "quota.mysql.col.num", &(state->quota_db_mysql_col_num),
      FLAGS_STRING, NULL},
#endif
#if WITH_MYSQLLOG
    { "mysqllog.host",     &(state->mysqllog_host),     FLAGS_STRING, NULL },
    { "mysqllog.user",     &(state->mysqllog_user),     FLAGS_STRING, NULL },
    { "mysqllog.password", &(state->mysqllog_password), FLAGS_STRING, NULL },
    { "mysqllog.database", &(state->mysqllog_database), FLAGS_STRING, NULL },
    { "mysqllog.query",    &(state->mysqllog_query),    FLAGS_STRING, NULL },
#endif
    { "mailinglists.file", &(state->mailinglists_file), FLAGS_STRING, NULL },
    { "include",           cstate, FLAGS_INCLUDE, NULL },
    { "av.type",                  &(state->av.type), 
      FLAGS_STRING, NULL },
    { "av.default.header",        &(state->av_default_header), 
      FLAGS_STRING, "X-Antivirus" },
    { "av.checkall",              &(state->av_checkall), FLAGS_BOOL, NULL },
    { "av.drweb.servers",         &(state->av.drweb.arg),
      FLAGS_ARGS, NULL },
    { "av.drweb.timeout",         &(state->av.drweb.timeout),
      FLAGS_TIME, NULL },
    { "av.drweb.servers.roll",    &(state->av.drweb.roll_percent),
      FLAGS_INT, NULL },
    { "av.sophie.servers",         &(state->av.sophie.arg),
      FLAGS_ARGS, NULL },
    { "av.sophie.timeout",         &(state->av.sophie.timeout),
      FLAGS_TIME, NULL },
    { "av.sophie.servers.roll",    &(state->av.sophie.roll_percent),
      FLAGS_INT, NULL },
    { "av.clamd.servers",         &(state->av.clamd.arg),
      FLAGS_ARGS, NULL },
    { "av.clamd.timeout",         &(state->av.clamd.timeout),
      FLAGS_TIME, NULL },
    { "av.clamd.servers.roll",    &(state->av.clamd.roll_percent),
      FLAGS_INT, NULL },
    { "av.clamd.mbox_format",     &(state->av.clamd_mbox_format),
      FLAGS_BOOL, NULL },
    { "av.virus.keeper",          &(state->av.viruskeeper.user), 
      FLAGS_STRING, NULL},
    { "av.action",                &(state->av.action),
      FLAGS_SLIST, NULL},
    { "spamassassin.servers",      &(state->spamassassin.arg),
      FLAGS_ARGS, NULL },
    { "spamassassin.timeout",      &(state->spamassassin.timeout),
      FLAGS_TIME, NULL },
    { "spamassassin.servers.roll", &(state->spamassassin.roll_percent),
      FLAGS_INT, NULL },
    { "spamassassin.acl",          &(state->spamassassin_acl), 
      FLAGS_SLIST, NULL },
    { "spamassassin.maxsize",      &(state->spamassassin_maxsize),
      FLAGS_INT, NULL },
    { "pipe.path",                 &(state->pipe_path), FLAGS_SLIST, NULL },
    { "filter.debug.msg",          &(state->filter_debug_msg),
      FLAGS_SLIST, NULL },
#if WITH_PERL
    { "perl",            &(state->perl) ,       FLAGS_BOOL,  NULL},
    { "perl.filtername", &(state->perl_filtername), FLAGS_STRING,
      PERL_FILTER},
    { "perl.eval", &(state->perl_eval), FLAGS_STRING, "filter", },
#endif
#ifdef UNSED_ACTION_LOG
    /* see log.c for comments */
    { "action.log.deliver.facility", &(state->action_log_deliver_facility),
      FLAGS_SYSLOG, NULL },
    { "action.log.forward.facility", &(state->action_log_forward_facility),
      FLAGS_SYSLOG, NULL },
    { "action.log.bounce.facility", &(state->action_log_bounce_facility),
      FLAGS_SYSLOG, NULL },
    { "action.log.drop.facility", &(state->action_log_drop_facility),
      FLAGS_SYSLOG, NULL },
    { "action.log.error.facility", &(state->action_log_error_facility),
      FLAGS_SYSLOG, NULL },
    { "action.log.deliver.format", &(state->action_log_deliver_format),
      FLAGS_STRING, NULL },
    { "action.log.forward.format", &(state->action_log_forward_format),
      FLAGS_STRING, NULL },
    { "action.log.bounce.format", &(state->action_log_bounce_format),
      FLAGS_STRING, NULL },
    { "action.log.drop.format", &(state->action_log_drop_format),
      FLAGS_STRING, NULL },
    { "action.log.error.format", &(state->action_log_error_format),
      FLAGS_STRING, NULL },
#endif
    { "filesystem.change.atime.when.write",
      &(state->filesystem_change_atime_when_write), FLAGS_BOOL, NULL },
    { "filesystem.with.noatime",
      &(state->filesystem_with_noatime), FLAGS_BOOL, NULL },
    { NULL, NULL, 0, NULL },
  };
  struct conf *spec_p=spec;
  if (cstate == NULL) 
    return NULL;
  while(spec_p->name) {
    if (!(spec_p->type & FLAGS_OBSOLETE))
      setval(spec_p->val, spec_p->type, spec_p->def);
    spec_p++;
  }
  spec_p=(struct conf *)malloc(sizeof(spec));
  if (spec_p == NULL)
    return NULL;
  cstate->level = 0;
  cstate->state = state;
  cstate->spec  = memcpy(spec_p, spec, sizeof(spec));
  return cstate;
}


static int readconf1(char *file, struct fstate *state, struct conf *spec) {
  FILE *f;
  char line[CONFLINEMAX];

  if ((f=fopen(file, "r"))==NULL) {
    syslog(LOG_ERR, "open %s: %m\n", file);
    return -1;
  }

  while(fgets(line, CONFLINEMAX, f)!=NULL) {
    char *xline = line;
    char *eline;
    char *args;
    struct conf *spec_p;

    while(isblank(*xline)) xline++;
    eline=strchr(xline, '\n');
      
    if (eline==NULL) {
      syslog(LOG_WARNING, "line too long in config file %s: %.20s", file, xline);
      fclose(f);
      return -1;
    }
    
    if (eline==line || *xline=='#')
      continue;
    /* skip ending spaces */
    eline--;
    while(eline>xline && isblank(*eline))
      eline--;
    if (eline==line)
      continue;
    eline[1]=0;

    args=xline;
    while(*args && !isblank(*args))
      args++;
    if (*args) {
      *args++=0;
      while(isblank(*args))
	args++;
    } else {
      *args=0;
    }
    
    spec_p=spec;
    while(spec_p->name) {
      if (strcasecmp(spec_p->name, xline)==0) {
	if (spec_p->type & FLAGS_OBSOLETE) {
	  syslog(LOG_WARNING, 
		 "in %s, configuration option '%s' is obsolete, use '%s' instead",
		 file, spec_p->name, spec_p->def);
	}
	if (setval(spec_p->val, spec_p->type & FLAGS_MASK, args)!=0) {
	  syslog(LOG_WARNING, "in %s, error while %s to '%s'",
		 file, spec_p->name, args);
	  fclose(f);
	  return -1;
	}
	break;
      }
      spec_p++;
    }
  }

  fclose(f);
  return 0;
}

int readconf(char *filename, struct fstate *state) {
  struct conf_state *cstate;
  int ret = -1;
  if ((cstate = resetconf(state)) != NULL) {
    ret = readconf1(filename, cstate->state, cstate->spec);
    free(cstate->spec);
  }
  return ret;
}

