/*
   Copyright 2000, 2001, 2002, 2003 Laurent Wacrenier

   This file is part of PLL lmtpd.

   PLL lmtpd is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   PLL lmtpd is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with PLL lmtpd; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include "config.h"

static char const rcsid[] UNUSED =
"$Id: quotawarn.c,v 1.24 2004/10/26 09:59:20 lwa Exp $";



#include <sys/types.h>
#include <sys/stat.h>
#include <sys/file.h>

#include <errno.h>
#include <syslog.h>
#include <fcntl.h>

#include <stdarg.h>
#include <stdio.h>
#include <time.h>
#include <string.h>
#include <ctype.h>

#include "lmtp.h"


int quotawarn(struct fstate *state, struct rcpt *rcpt1) {
  char domain[255];
  char tmpname[PATH_MAX];
  FILE *msg, *tmp;
  int c;
  int markfd;
  struct stat sb;

  if (state->quota_mark==NULL || *state->quota_mark==0)
    return -1;

  if (stat(state->quota_mark, &sb)==-1) {
    if (errno!=ENOENT)
      return 0;
  } else {
    int delay = state->date - sb.st_mtime;
    if (delay < 0)
      delay = 0;
    return delay;
  }

  if ((markfd=open(state->quota_mark, O_WRONLY|O_CREAT|O_TRUNC, 0644))==-1)
    return 0; 
  if (close(markfd)==-1)
    return 0;

  if (state->quota_warnmsg==NULL || *state->quota_warnmsg==0)
    return 0;


  getuserdomain(state->qrecipient, domain);

  if ((msg = opentemplate(state->quota_warnmsg, domain))==NULL)
    return 0;

  if ((tmp=newmessage(state, "quota", tmpname))==NULL) {
    fclose(msg);
    return 0;
  }

  while((c=fgetc(msg))!=EOF) {
    if (c=='%') {
      c=fgetc(msg);
      switch(c) {
      case 'D':
	fputs(domain, tmp);
	break;
      case 'R':
	fputs(state->qrecipient, tmp);
	break;
      case 'Q':
	fprintf(tmp, "%lu", rcpt1->quota);
	break;
      case EOF:
	break;
      default:
	fputc(c, tmp);
      }
    } else {
      if (c != '\r') /* remove this junk */
	fputc(c, tmp);
    }
  }
  fclose(msg);

  if (endmessage(state, "quota", tmpname, tmp)==EOF) {
    return 0;
  }
  syslog(LOG_INFO, "overquota message sent to <%s>", 
	 state->qrecipient);
  return 0;
}


static void vsyslog_warning(const char *fmt, va_list args) {
  vsyslog(LOG_WARNING, fmt, args);
}


static void quotadb_null_add(struct fstate *state, char *user, int code) {}
static void quotadb_null_del(struct fstate *state, char *user) {}
static char *quotadb_null_check(struct fstate *state, char *user) {
  return NULL;
}
static void quotadb_null_close(struct fstate *state) {}

void (*quotadb_add)(struct fstate *state, char *user, int code) 
     = quotadb_null_add ;
     
void (*quotadb_del)(struct fstate *state, char *user) 
     = quotadb_null_del ;

char *(*quotadb_check)(struct fstate *state, char *user)
     = quotadb_null_check ;

void (*quotadb_close)(struct fstate *state) 
     = quotadb_null_close;

     
int quotadb_open(struct fstate *state) {
  int ret;

  mdq_error=vsyslog_warning;

  if (state->quota_db_timeout<=0) 
    return 0;

  if (state->quota_db_method == NULL)
    return 0;

  if (strcasecmp(state->quota_db_method, "link")==0) {
    if ( (ret=quotadb_link_open(state)) == 0) {
      quotadb_add = quotadb_link_add;
      quotadb_del = quotadb_link_del;
      quotadb_check = quotadb_link_check;
      quotadb_close = quotadb_link_close;
      
    }
    return ret;
  }
#if WITH_DB
  if (strcasecmp(state->quota_db_method, "db")==0 ||
      strcasecmp(state->quota_db_method, "db3")==0) { /* depreciated */
    if ( (ret=quotadb_db3_open(state)) == 0) {
      quotadb_add = quotadb_db3_add;
      quotadb_del = quotadb_db3_del;
      quotadb_check = quotadb_db3_check;
      quotadb_close = quotadb_db3_close;
    }
    return ret;
  }
#endif
#if WITH_MYSQLLOG
  if (strcasecmp(state->quota_db_method, "mysql")==0) {
    if ( (ret=quotadb_mysql_open(state)) == 0) {
      quotadb_add = quotadb_mysql_add;
      quotadb_del = quotadb_mysql_del;
      quotadb_check = quotadb_mysql_check;
      quotadb_close = quotadb_mysql_close;
    }
    return ret;
  }
  if (strcasecmp(state->quota_db_method, "mysql.add")==0) {
    if ( (ret=quotadb_mysqladd_open(state)) == 0) {
      quotadb_add = quotadb_mysqladd_add;
      quotadb_del = quotadb_null_del;
      quotadb_check = quotadb_null_check;
      quotadb_close = quotadb_mysql_close;
    }
    return ret;
  }
#endif
  return 0;
}

