/*
   Copyright 2000, 2001, 2002, 2003 Laurent Wacrenier

   This file is part of PLL lmtpd.

   PLL lmtpd is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   PLL lmtpd is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with PLL lmtpd; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include "config.h"

static char const rcsid[] UNUSED =
"$Id: regrep.c,v 1.9 2004/10/26 09:59:21 lwa Exp $";

#include "../config.h"

#include <sys/types.h>
#include <sys/time.h>
#include <sys/resource.h>

#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "re.h"

void usage(int ret) {
  fprintf(stderr, "usage: regrep [-n NN] pattern\n");
  exit(ret);
}

double timevald(struct timeval *a) {
  return (double)a->tv_sec + (double)a->tv_usec/1000000;
}

RE *compile(char *pattern, int nrun, int closeit) {
  char **srcptr;
  struct rusage ru_before, ru_after;
  RE *re = NULL;
  int n = nrun;
  char *xpat;
  char *err;

  if ( (err = re_init()) != NULL) {
    fprintf(stderr, "%s", err);
    exit(1);
  }

  printf("compiling %s\n", pattern);
  if (nrun)
    getrusage(0, &ru_before);
  do {
    xpat = pattern;
    if (closeit && re) 
      re->close(re);
    re = re_find(&xpat);
  } while (--n>0);
  if (nrun) {
    getrusage(0, &ru_after);
    printf("compile time*%d: user %.2f system %.2f\n", nrun,
	   timevald(&ru_after.ru_utime) -  timevald(&ru_before.ru_utime),
	   timevald(&ru_after.ru_stime) -  timevald(&ru_before.ru_stime));
  }
  if (*xpat)
    printf("remaining '%s'\n", xpat);
  if (re->hint) {
    printf("has hint\n");
  }
  switch(re->err) {
  case RE_ERR_NOERR: 
    printf("compile ok\n"); break;
  case RE_ERR_MEMORY:
    printf("compile memory error (%s)\n", re->errstr); break;
  case RE_ERR_SYNTAX:
    printf("compile syntax error (%s)\n", re->errstr); break;
  }
  return re;
}

int main(int argc, char *argv[]) {
  RE *re;
  char ch;
  char buf[1024];
  char **srcptr;
  int nrun = 0;
  int i;
  int closeit = 1;

  while((ch = getopt(argc, argv, "ckn:"))!=-1) {
    switch (ch) {
    case 'n':
      nrun = atoi(optarg); break;
    case 'k':
      closeit = 0; break;
#if WITH_REGEXP_DB_CACHE
    case 'c':
      re_cache_init(); break;
#endif
    case '?':
    default:
      usage(1);
    }
  }
  argc -= optind;
  argv += optind;
  if (argv[0]==NULL)  usage(1);

  for (i = 0; i<argc; i++) {
    re = compile(argv[i], nrun, closeit);
  }

  while(fgets(buf, 1023, stdin)!=NULL) {
    char *end = strchr(buf, '\n');
    int ret;
    int n = nrun;
    struct rusage ru_before, ru_after;
    if (end) *end = 0;
    
    if (*buf == '!') {
      if (re)
	re->close(re);
      re = compile(buf + 1, nrun, closeit);
      continue;
    }

    if (nrun)
      getrusage(0, &ru_before);
    do {
      ret = re->exec(re, buf);
    } while(--n>0);
    if (nrun) {
      getrusage(0, &ru_after);
      printf("run time*%d: user %.2f system %.2f\n", nrun,
	     timevald(&ru_after.ru_utime) -  timevald(&ru_before.ru_utime),
	     timevald(&ru_after.ru_stime) -  timevald(&ru_before.ru_stime));
    }
    printf("==> %d ", ret);
    switch(ret) {
    case 0: printf("match"); break;
    case 1: printf("not match"); break;
    default: printf("error (%s)", re->errstr); break;
    }
    printf("\n");
    
  }
  re->close(re);
  exit(0);
}
