/*
   Copyright 2000, 2001, 2002, 2003 Laurent Wacrenier

   This file is part of PLL lmtpd.

   PLL lmtpd is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   PLL lmtpd is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with PLL lmtpd; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include "config.h"

static char const rcsid[] UNUSED =
"$Id: rfc2047.c,v 1.2 2005/08/03 11:04:46 lwa Exp $";



#ifdef DMALLOC
#include <dmalloc.h>
#endif

#include <sys/types.h>

#include <syslog.h>
#if WITH_ICONV
#include <iconv.h>
#endif
#include <regex.h>

#include <string.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>

#include "rfc.h"

/* I'm too lazy to parse it myself */
/* TODO: write a real parser */

#define TOKEN "[a-zA-Z0-9_+*$#'{}-]+"
#define RFC2047_SYNTAX "=\\?(" TOKEN ")\\?(" TOKEN \
                       ")\\?([\x21-\x3e\x40-\x7e]+)\\?="


#if RFC2047_C_MAIN
#include <stdio.h> 
#define DEBUG(x...) fprintf(stderr, ##x)
#else
#define DEBUG(x...)
#endif


static regex_t rfc2047_re;

int rfc2047_init(void) {
  int err;
  if ((err = regcomp(&rfc2047_re, RFC2047_SYNTAX, REG_EXTENDED)) != 0) {
    char errbuf[100];
    regerror(err, &rfc2047_re, errbuf, sizeof errbuf);
    syslog(LOG_ERR, "rfc2047_init(%s): %s\n", RFC2047_SYNTAX, errbuf);
    return -1;
  }
  return 0;
}

/* step for memory allocation */
#define STEP 8192
#define TOCODE "UTF-8"

static char *utf8(char *s, size_t len, char *charset) {
#if WITH_ICONV
  char *mem = NULL;
  iconv_t di = iconv_open(TOCODE, charset);
  size_t memlen = 0;
  char *outbuf;
  size_t outbytesleft;
#if ICONV_CONST
  const char *inbuf; /* non POSIX iconv (ex: libiconv) */
#else
  char *inbuf;
#endif
  size_t inbytesleft;
  if (di == (iconv_t)-1) {
    DEBUG("iconv_open(%s): %s\n", charset, strerror(errno));
    syslog(LOG_ERR, "iconv_open(%s,%s): %s", charset, TOCODE, strerror(errno));
    return NULL;
  }
  inbuf = s;
  inbytesleft = len;
  memlen += STEP;
  mem = realloc(mem, memlen);
  outbuf = mem;
  outbytesleft = memlen;
  while(1) {
    size_t err;
    if ((err = iconv(di, &inbuf, &inbytesleft, &outbuf, &outbytesleft)) 
	== (size_t)-1) {
      DEBUG("iconv: %s", strerror(errno));
      DEBUG("inbuf=%.*s\n", (int)inbytesleft, inbuf);
      if (errno == E2BIG) {
	char *omem = mem;
	memlen += STEP;
	mem = realloc(omem, memlen+1);
	if (mem == NULL) {
	  iconv_close(di);
	  free(omem);
	  return NULL;
	}
	outbuf += mem - omem;
	outbytesleft = memlen - (outbuf - mem);
      } else {
	inbuf ++;
	inbytesleft --;
      }
    } else {
      break;
    }
  }
  *outbuf = 0;
  iconv_close(di);
  return mem;
#else
  errno = 0;
  return NULL;
#endif
}

static int hexdigit(char h) {
  switch(h) {
  case '0': return 0;
  case '1': return 1;
  case '2': return 2;
  case '3': return 3;
  case '4': return 4;
  case '5': return 5;
  case '6': return 6;
  case '7': return 7;
  case '8': return 8;
  case '9': return 9;
  case 'a': case 'A': return 0xA;
  case 'b': case 'B': return 0xB;
  case 'c': case 'C': return 0xC;
  case 'd': case 'D': return 0xD;
  case 'e': case 'E': return 0xE;
  case 'f': case 'F': return 0xF;
  default: return -1;
  }
}

static size_t decode_xqp(char *s, char *d, ssize_t len) {
  int olen = 0;
  while(len > 0) {
    switch(*s) {
    case '_':
      *d++ = 0x20;
      s++;
      len --;
      olen ++;
      break;
    case '=':
      if (len >= 3 ) {
	int c1 = hexdigit(s[1]);
	int c2 = hexdigit(s[2]);
	if (c1 >= 0 && c2 >= 0) {
	  *d++ = c1 * 16 + c2;
	  olen ++;
	}
      }
      s+=3;
      len -= 3;
      break;
    default:
      *d++ = *s++;
      len --;
      olen ++;
      break;
    }
  }
  return olen;
}

static char *add_to_string(char **buf_p, size_t *buflen_p, char **curs_p, 
		    char *s, size_t slen) {
  if (*buf_p == NULL || *curs_p - *buf_p + slen > *buflen_p) {
    char *nbuf = *buf_p;
    *buflen_p += slen + 1024;
    nbuf = realloc(*buf_p, *buflen_p);
    if (nbuf == NULL) {
      free(*buf_p);
      *buf_p = NULL;
      *curs_p = NULL;
      *buflen_p = 0;
      return NULL;
    }
    if (*buf_p == NULL)
      *curs_p = *buf_p;
    *curs_p += (nbuf - *buf_p);
    *buf_p = nbuf;
  }
  if (s) {
    DEBUG("copy %p %p -> %s\n", *buf_p, *curs_p, s);
    memcpy(*curs_p, s, slen);
    *curs_p += slen;
  }
  return *curs_p;
}

char *rfc2047_read(char *start, char *convert_from) {
  size_t len = strlen(start);
  char *curs = convert_from;
  char *dest = NULL;
  char *buf = malloc(len+1); /* encoding is memory hungry */
  char *dest_c = NULL;
  size_t dest_len = 0;
  int nbloc = 0;

  add_to_string(&dest, &dest_len, &dest_c, start, convert_from - start);
  
  do {
    regmatch_t pmatch[4];
    if (regexec(&rfc2047_re, curs, 4, pmatch, 0) == 0) {
      char charset[100];
      int charset_len = pmatch[1].rm_eo-pmatch[1].rm_so;
      if (charset_len >= sizeof(charset))
	charset_len = sizeof(charset);
      memcpy(charset, curs + pmatch[1].rm_so, charset_len);
      charset[charset_len] = 0;

      if (pmatch[0].rm_so) {
	char *s = curs;
	/* linear-space-white between encoded-word are ignored */
	if (nbloc > 0)
	  while(*s == ' ' || *s == '\t' || *s == '\n')
	    s++;
	DEBUG("RAW %.*s\n", (int)pmatch[0].rm_so, curs);
	if (s < curs + pmatch[0].rm_so) {
	  if (add_to_string(&dest, &dest_len, &dest_c, curs, pmatch[0].rm_so) 
	      == NULL) {
	    free(buf);
	    return NULL;
	  }
	}
      }

      DEBUG("RFC2047 charset=%s encoding=%.*s data=%.*s\n",
	    charset,
	    (int)(pmatch[2].rm_eo-pmatch[2].rm_so), curs + pmatch[2].rm_so, 
	    (int)(pmatch[3].rm_eo-pmatch[3].rm_so), curs + pmatch[3].rm_so);
      if (pmatch[2].rm_eo-pmatch[2].rm_so == 1) {
	char *u = NULL;
	size_t len;
	if (curs[pmatch[2].rm_so] == 'Q' || 
	    curs[pmatch[2].rm_so] == 'q') {
	  len = decode_xqp(curs + pmatch[3].rm_so, buf, 
			   pmatch[3].rm_eo-pmatch[3].rm_so);
	  DEBUG("QP %.*s\n", len, buf);
	} else if (curs[pmatch[2].rm_so] == 'B' || 
		   curs[pmatch[2].rm_so] == 'b') {
	  len = base64_decode(curs + pmatch[3].rm_so,
			      pmatch[3].rm_eo-pmatch[3].rm_so,
			      (unsigned char *)buf);
	  if (len == (size_t)-1) { /* undecodable string */
	    DEBUG("base64: %s", strerror(errno));
	    add_to_string(&dest, &dest_len, &dest_c, 
			  curs + pmatch[3].rm_so,
			  pmatch[3].rm_eo-pmatch[3].rm_so);
	    curs += pmatch[0].rm_eo;
	    continue;
	  }
	  DEBUG("B64 %.*s\n", (int)len, buf);
	} else {
	  if (add_to_string(&dest, &dest_len, &dest_c, curs + pmatch[3].rm_so,
			    pmatch[3].rm_eo-pmatch[3].rm_so) == NULL) {
	    free(buf);
	    return NULL;
	  }
	  len = 0;
	}
	if (len) { /* something to convert */
	  u = utf8(buf, len, charset);
	  if (u == NULL && errno == ENOMEM)
	    return NULL;
	  if (u == NULL) {
	    if (add_to_string(&dest, &dest_len, &dest_c, buf, len) == NULL) {
	      free(buf);
	      return NULL;
	    }
	  } else {
	    DEBUG("UTF-8 %s\n", u);
	    if (add_to_string(&dest, &dest_len, &dest_c, u, strlen(u))
		== NULL) {
	      free(u);
	      free(buf);
	      return NULL;
	    }
	    free(u);
	  }
	}
      }
      curs += pmatch[0].rm_eo;
      nbloc ++;
    } else {
      DEBUG("END %s", curs);
      if (add_to_string(&dest, &dest_len, &dest_c, curs, strlen(curs)) 
	  == NULL) {
	free(buf);
	return NULL;
      }
      break;
    }
  } while(1);
  *dest_c++ = 0;
  dest = realloc(dest, dest_c - dest); /* free extra memory */
  DEBUG("FINAL: %s\n", dest);
  free(buf);
  return dest;
}

#ifdef RFC2047_C_MAIN
int main(int argc, char **argv) {
  char buf[1024];
  rfc2047_init();
  while(fgets(buf, sizeof(buf), stdin)) {
    rfc2047_read(buf, buf);
  }
}

#endif
