/*
   Copyright 2000, 2001, 2002, 2003 Laurent Wacrenier

   This file is part of PLL lmtpd.

   PLL lmtpd is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   PLL lmtpd is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with PLL lmtpd; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include "config.h"

static char const rcsid[] UNUSED =
"$Id: send.c,v 1.67 2005/02/22 12:12:57 lwa Exp $";



#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <sys/uio.h>
#include <sys/resource.h>

#include <regex.h>
#include <syslog.h>
#include <sysexits.h>
#include <fcntl.h>
#include <unistd.h>

#if HAVE_GRP_H
#include <grp.h>
#endif

#include <errno.h>
#include <limits.h>
#include <ctype.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "lmtp.h"

#define BUFLEN 65536


#define MD_NEW "new"
#define MD_CUR "cur"
#define TRASH "Trash"

/* Fix for FreeBSD 4 and other bogus systems */
#if !IOV_MAX
#define IOV_MAX 16
#endif

/*
 *  set error according sysexits
 */

static char *sysexits(struct fstate *state, int err) {
  switch(err) {
  case EX_OK:
    OkNoError(state);
    return NULL;
  case EX_USAGE:
    return seterr(state, 500, "5.5.4", "Bad usage");
  case EX_DATAERR:
    return seterr(state, 501, "5.5.2", "Data format error");
  case EX_NOUSER:
    return seterr(state, 550, "5.1.1", "User unknown");
  case EX_NOHOST:
    return seterr(state, 550, "5.1.2", "Host unknown");
  case EX_NOINPUT:
    return seterr(state, 550, "5.3.0", "Cannot open input");
  case EX_CANTCREAT:
    return seterr(state, 550, "5.3.0", "Can't create output file");
  case EX_NOPERM:
    return seterr(state, 550, "5.3.0", "Insufficient permission");
  case EX_UNAVAILABLE:
    return seterr(state, 554, "5.5.0", "Service unavailable");
  case EX_SOFTWARE:
    return seterr(state, 554, "5.5.0", "Internal error");
  case EX_OSFILE:
    return seterr(state, 554, "5.5.0", "System file missing");
  case EX_PROTOCOL:
    return seterr(state, 554, "5.5.0", "Remote protocol error");
#ifdef EX_CONFIG
    /* not defined on HPUX */
  case EX_CONFIG:
    return seterr(state, 554, "5.5.0", "Local configuration error");
#endif
  case EX_OSERR:
    return seterr(state, 451, "4.5.0", "System error");
  case EX_IOERR:
    return seterr(state, 451, "4.5.0", "Input/output error");
  case EX_TEMPFAIL:
    return seterr(state, 450, "4.2.0", "Deferred");  /* ??? */
  default:
    return seterr(state, 550, "5.5.4", "unknown error %d", err);
  }
}

/*
 * write mail headers and body into a file descriptor
 * return number of wroten bytes
 */

ssize_t mailwrite(struct fstate *state, struct headers *hdrlist,
		  int level, int fd) {
  struct iovec *iov;
  struct iovec *iov_c, *iov_e;
  ssize_t wrlen = 0;
  ssize_t tolen = 0;

  iov = malloc((2*state->nbheaders+1) * sizeof (struct iovec));
  
  if (iov == NULL) {
    MemoryError(state);
    return -1;
  }
  iov_c = iov;

  while(hdrlist) {
    if (hdrlist->type <= level) {
      tolen+=hdrlist->len+1;
      iov_c->iov_len  = hdrlist->len;
      iov_c->iov_base = hdrlist->header;
      iov_c++;
      iov_c->iov_len  = 1;
      iov_c->iov_base = "\n";
      iov_c++;
    }
    hdrlist=hdrlist->next;
  }

  if (state->bodysize > 0) {
    iov_c->iov_len  = state->bodysize;
    iov_c->iov_base = state->body;
    iov_c ++;
    tolen += state->bodysize;
  }

  iov_e = iov_c; /* ending */
  iov_c = iov; /* cursor */
  while(iov_c < iov_e) {
    int niov = iov_e - iov_c;
    int len;

    if (niov > IOV_MAX) 
      niov = IOV_MAX;
    len=writev(fd, iov_c, niov);
    if (len < 0) {
      DeliveryError(state, "writev error");
      free(iov);
      return -1;
    }
    wrlen += len;
    iov_c += niov;
  }

  if (wrlen != tolen) {
    seterr(state, 451, "4.2.0", "bad write count (%lu!=%lu)",
	   (unsigned long)tolen, (unsigned long)wrlen);
  }
  free(iov);

  if (fsync(fd)==-1 && errno != EINVAL) { /* ok if socket */
    DeliverySysError(state, "fsync");
    return -1;
  }

  return wrlen;
}

/* returns the size of the mail according the previous routine */
ssize_t mailwrite_size(struct fstate *state, struct headers *hdrlist,
		       int level) {
  ssize_t headersize = 0;
  while(hdrlist) {
    if (hdrlist->type <= level)
      headersize += hdrlist->len+1;
    hdrlist = hdrlist->next;
  }
  return headersize + state->bodysize;
}


static char *sendmd_final(struct fstate *state, char *box, 
		    char *tmpname, ssize_t total_size, char *new) {
  int already_delivered;
  struct stat sb;
  char newname[PATH_MAX];

  /* see http://cr.yp.to/proto/maildir.html for format */

  snprintf(newname, PATH_MAX, ".%s/%s/%lu.P%luQ%lu.%s,S=%lu",
	   getbox(box), new,
	   (unsigned long)state->date, (unsigned long)state->pid,
	   state->counter, state->myhostname, (unsigned long)total_size);
  /*
   * Diffrences with standard Maildir
   *   http://www.qmail.org/man/man5/maildir.html
   * - no 24 hour timer or retry : delivery shall be fast,
   *   the retry in on the charge of LMTP client
   * - use rename instead of link/unlink (some
   *   maildir reader incorrectly use link count)
   */
  
  already_delivered = (stat(newname, &sb) == 0); /* trace for soft quota */

  if (rename(tmpname, newname)==-1) {
    if (errno == ENOENT) {
      mkbox(state, box);
      if (rename(tmpname, newname)==-1) {
	unlink(tmpname);
	return "rename failed";
      }
    } else {
      unlink(tmpname);
      return  "rename failed";
    }
  }
  if (!already_delivered && state->quotainfo && strcmp(box, TRASH) != 0)
    mdq_add(state->quotainfo, (long)total_size, +1);
  return NULL;
}

/*
 * deliver to an maildir
 */

char *sendmd(struct fstate *state, char *box) {
  char tmpname[PATH_MAX];
  int tmpfd;
  ssize_t total_size;
  char *new=MD_NEW;
  struct headers *output_headers=state->hdrlist;
  char *err;

  if ((tmpfd=opentmp(state, tmpname))==-1)
    return DeliveryError(state, "unable to open user temporary file");

  getbox(box);

  if (state->rcpt->shared) {
    fchmod(tmpfd, 0444);  /* protect the directories instead */
    new=MD_CUR;
    if (box && *box)             /* hide recipients traces */
      output_headers=state->hdr_delivered->next;
  }
  
  if ((total_size=mailwrite(state, output_headers, state->hlevel, tmpfd))<0) {
    unlink(tmpname);
    return state->error;
  }

  if (close(tmpfd)<0) {
    unlink(tmpname);
    return DeliverySysError(state, "close");
  }

  err = sendmd_final(state, box, tmpname, total_size, new);
  if (err) 
    return DeliverySysError(state, err);
  else
    return NULL;
}


/*
 * start sendmail
 * return a file descriptor for input socket or -1 on error
 */

int runsendmail(struct fstate *state,
		char *sendmail, char *sender, char *destination) {
  int pp[2];
  char * volatile x_sender=NULL, * volatile x_destination=NULL;

  if (*destination) {
    char *x_destination1=quote_addr821(destination);
    if (x_destination1) {
      if ((x_destination=malloc(3+strlen(x_destination1)))!=NULL)
	sprintf(x_destination, "<%s>", x_destination1);
      free(x_destination1);
    }
  } else {
    seterr(state, 551, "5.5.0", "user not local; no specified destination");
    return -1;
  }
  
  if (*sender) {
    char *x_sender1=quote_addr821(sender);
    if (x_sender1) {
      if ((x_sender=malloc(3+strlen(x_sender1)))!=NULL)
	sprintf(x_sender, "<%s>", x_sender1);
      free(x_sender1);
    }
  } else {
    x_sender=strdup("<>");
  }

  if (x_sender==NULL || x_destination==NULL) {
    free(x_sender); free(x_destination);
    MemoryError(state);
    return -1;
  }

  if (pipe(pp) < 0) {
    free(x_sender); free(x_destination);
    seterr(state, 451, "4.5.0", "pipe: %s", strerror(errno));
    return -1;
  }

  switch (vfork()) {
  case -1:
    free(x_sender); free(x_destination);
    close(pp[0]);
    close(pp[1]);
    seterr(state, 451, "4.5.0", "unable to fork : %s", strerror(errno));
    return -1;

  case 0:
    close(pp[1]);
    if (pp[0] != 0) {
      dup2(pp[0], 0);
      close(pp[0]);
    }
    close(2);
    execl(sendmail, "sendmail", "-i", "-f",
	  x_sender, x_destination, NULL);
    syslog(LOG_ERR, "Unable to run sendmail, %s: %s",
	   sendmail, strerror(errno));
    _exit(EX_OSERR);
  }

  free(x_sender); free(x_destination);
  close(pp[0]);
  return pp[1];
}

/*
 * wait for sendmail exits
 */

char *waitsendmail(struct fstate *state, int fd) {
  int status;

  if (fd != -1 && close(fd) == -1) {
      DeliveryError(state, "close");
      fd = -1;
  }
  if (wait(&status)==-1) {
    if (fd==-1) 
      return seterr(state, 450, "4.5.0", "wait");
  } else {
    if (fd==-1)
      return state ? state->error : NULL;
    if (WIFEXITED(status)) {
      return sysexits(state, WEXITSTATUS(status));
    } else {
      return seterr(state, 450, "4.5.0", "sendmail not properly exited");
    }
  }
  return NULL; /* never happen */
}


/* 
 *  send a mail with sendmail
 *  returns NULL if ok
 *  else returns error string 
 */

char *sendmail(struct fstate *state, char *sender, char *destination) {
  int fd;

  fd = runsendmail(state, state->sendmail_path, sender, destination);
  if (fd == -1)
    return state->error;

  if ( mailwrite(state, state->hdr_delivered, state->hlevel, fd) < 0 ) {
    close(fd);
    fd=-1;
  }
  
  return waitsendmail(state, fd);
}


/*
 * check if name if in the vacation cache
 */

static int vcache(struct fstate *state, char *name) {
  FILE *f;
  char buf[1024];
  const char *openmode = state->vacationcache_create ? "a+" : "r+";
  if (state->vacationcache==NULL)
    return -1;
  if ((f=fopen(state->vacationcache, openmode))==NULL)
    return -1;
  if (state->vacationcache_create)
    rewind(f);
  while(fgets(buf, sizeof(buf), f)!=NULL) {
    char *end;
    if (buf[0]=='\n') {
      fseek(f, -1L, SEEK_CUR);
      break;
    }
    end=strchr(buf, '\n');
    if (end!=0)
      *end=0;
    fprintf(stderr, "cmp '%s' '%s'\n", name, buf);
    if(strcasecmp(name, buf)==0) {
      fclose(f);
      return -1;
    }
  }
  fputs(name, f);
  fputs("\n\n", f);
  fclose(f);
  return 0;
}

/*
 * send vacation if any
 * return a message to happend at log line.
 */

char *send_vacation(struct fstate *state, char *message) {
  int fd;
  FILE *f;
  struct headers *hl;
  static regex_t preg;
  static int preg_done=0;
  int ret;

  if (message==NULL)
    return "";
  if (state->sender==NULL || state->sender[0]==0)
    return "";
  if (!preg_done) {
    if ((ret=regcomp(&preg, FROMDsubstitute,
		     REG_EXTENDED|REG_ICASE|REG_NOSUB))!=0) {
      char err[1024];
      regerror(ret, &preg, err, sizeof(err)-1);
      syslog(LOG_ERR, "unable to compile daemon discovery regex: %s", err);
      return "";
    }
    preg_done=1;
  } 
  for (hl=state->hdrlist; hl; hl=hl->next) {
    if (regexec(&preg, hl->header, 0, NULL, 0)==0) {
      return "";
    }
  }

  if (vcache(state, state->sender)!=0)
    return "";

  fd=runsendmail(NULL, state->sendmail_path, "", state->sender);

  if (fd==-1)
    return "";

  if ((f=fdopen(fd, "a"))!=NULL) {
    char *subject=NULL;

    for (hl=state->hdrlist; hl; hl=hl->next) {
      if (hl->type == HEADER_DELIVERED_TO || hl->type == HEADER_RECEIVED) {
	fprintf(f, "%s\n", hl->header);
      } else if (!subject && hl->type == HEADER_SUBJECT) {
	subject=hl->header+hl->dd_len+1;
	while(isspace((int)*subject))
	  subject++;
	if (tolower(subject[0])=='r' &&
	    tolower(subject[1])=='e' &&
	    subject[2]==':') {
	  subject+=3;
	  while(isspace((int)*subject))
	    subject++;
	}
      }
    }
    fprintf(f, "References: %s\n", state->messageid);
    fprintf(f, "From: %s\n", state->rcpt->user);
    fprintf(f, "To: %s\n", state->sender);
    if (state->charset) 
      fprintf(f, "Content-Type: text/plain; charset=%s\n", state->charset);
    if (subject)
      fprintf(f, "Subject: Re: %s\n", subject);
    fprintf(f, "In-Reply-To: %s\n", state->messageid);
    fprintf(f, "Auto-Submitted: auto-replied\n"); /* RFC 3834 */
    fprintf(f, "\n");
    fputs(message, f);
    fclose(f);
  } else {
    close(fd);
  }
  waitsendmail(NULL, fd);
  return " (vacation message send)";
}

char *send_pipe(struct fstate *state, char *box, char *argv[], char *path) {
  int p[2];
  pid_t child;
  char in_path[PATH_MAX];
  char out_path[PATH_MAX];
  int in_fd;
  int out_fd;
  FILE *out_f;
  char buf[8192];
  size_t out_len = 0;
  size_t in_len ;
  ssize_t read_len;
  int status;

  if (strchr(argv[0], '/') != NULL)
    return "invalid char in command";

  if ( (in_fd = opentmp(state, in_path)) == -1)
    return "unable to open input file";

  if ((in_len = mailwrite(state, state->hdr_my_received->next,
			  state->hlevel, in_fd))<0) {
    unlink(in_path);
    return state->error;
  }

  if ((in_fd = open(in_path, O_RDONLY)) == -1) {
    unlink(in_path);
    return "unable to reopen input file";
  }
  unlink(in_path);

  if ( (out_fd = opentmp(state, out_path)) == -1) {
    close(in_fd);
    return "unable to open output file";
  }

  if (pipe(p) == -1) {
    unlink(out_path);
    close(in_fd);
    close(out_fd);
    return "pipe failed";
  }

  switch(child = fork()) {
  case -1:
    unlink(out_path);
    close(in_fd);
    close(out_fd);
    close(p[0]);
    close(p[1]);
    return "fork failed";

  case 0: {
    struct rlimit rl;
    rlim_t maxfd;
    extern char **environ;
    char *env[100];
    char env_string[ARG_MAX];
    char *env_string_c = env_string;
    char *env_string_end = env_string+sizeof(env_string);
    int env_n = 0;
    char *at = strrchr(state->rcpt->rcpt, '@');


#if HAVE_SETGROUPS
    setuid(0);
    if (setgroups(0, NULL) == -1)
      _exit(EX_NOPERM);
#endif

    if (setgid(state->rcpt->gid) == -1)
      _exit(EX_NOPERM);
    if (setuid(state->rcpt->uid) == -1)
      _exit(EX_NOPERM);

    close(out_fd);
    close(p[0]);
    if (p[1] != 1) {
      dup2(p[1], 1);
      close(p[1]);
    }
    if (in_fd != 0) {
      dup2(in_fd, 0);
      close(in_fd);
    }
    /*     close(2); */ /* TODO: manage stderr */

    /* close all files */
    if (getrlimit(RLIMIT_NOFILE, &rl) != -1) {
      maxfd = rl.rlim_cur;
      if (maxfd == RLIM_INFINITY 
#ifdef RLIM_SAVED_MAX
	  || maxfd == RLIM_SAVED_MAX
#endif
#ifdef RLIM_SAVED_CUR
	  || maxfd == RLIM_SAVED_CUR
#endif
	  )
	maxfd = 64; /* expect it's enough */
    } else {
      maxfd = 64;
    }
    
    while(maxfd > 2)
      close(maxfd--);

    env[env_n++] = env_string_c;
    if ( env_string_end > env_string_c ) {
      env_string_c += snprintf(env_string_c, env_string_end - env_string_c,
			       "PATH=%s", path) + 1;
    }
    env[env_n++] = env_string_c;
    if ( env_string_end > env_string_c ) {
      env_string_c += snprintf(env_string_c, env_string_end - env_string_c,
			       "RECIPIENT=%s", state->rcpt->rcpt) + 1;
    }
    env[env_n++] = env_string_c;
    if ( env_string_end > env_string_c ) {
      env_string_c += snprintf(env_string_c, env_string_end - env_string_c,
			       "MAILBOX=%s", state->rcpt->mbox) + 1;
    }
    env[env_n++] = env_string_c;
    if ( env_string_end > env_string_c ) {
      env_string_c += snprintf(env_string_c, env_string_end - env_string_c,
			       "HOME=%s", state->rcpt->path) + 1;
    }
    if (state->rcpt->ext) {
      env[env_n++] = env_string_c;
      if ( env_string_end > env_string_c ) {
	env_string_c += snprintf(env_string_c, env_string_end - env_string_c,
				 "EXTENSION=%s", state->rcpt->ext) + 1;
      }
    }
    if (at) {
      env[env_n++] = env_string_c;
      if ( env_string_end > env_string_c ) {
	env_string_c += snprintf(env_string_c, env_string_end - env_string_c,
				 "DOMAIN=%s", at+1) + 1;
      }
      env[env_n++] = env_string_c;
      if ( env_string_end > env_string_c ) {
	env_string_c += snprintf(env_string_c, env_string_end - env_string_c,
				 "LOCAL=%.*s", at - state->rcpt->rcpt,
				 state->rcpt->rcpt) + 1;
      }
    }
    
    environ = env;

    execvp(argv[0], argv);
    _exit(EX_UNAVAILABLE);
  }
  }
  close(in_fd);
  close(p[1]);

  out_f = fdopen(out_fd, "w");

  out_len+=fprintf(out_f, "%s\n", state->hdrlist->header); /* Return-Path */
  out_len+=fprintf(out_f, "%s\n", state->hdr_delivered->header);
  out_len+=fprintf(out_f,
		   "Received: by %s (%s) with LMTP%s via PIPE\n"
		   "\tid %lu for <%s>\n",
		   state->myhostname, state->myname,
		   state->sasl_user ? "A" : "",
		   (unsigned long)child,state->qrecipient);
  out_len+=fprintf(out_f, "%s\n", state->hdr_my_received->header);
  
  while((read_len = read(p[0], buf, sizeof(buf))) > 0) {
    if (fwrite(buf, 1, read_len, out_f) < 0)
      break;
    out_len += read_len;
  }
  if (fclose(out_f) == EOF)
    read_len = 1; /* write error */

  close(p[0]);
  if (wait(&status) == -1) {
    unlink(out_path);
    return "wait failed";
  }

  switch(read_len) {
  case -1:
    unlink(out_path);
    return "read error";
  default:
    unlink(out_path);
    return "write error";
  case 0:
    break;
  }

  if (WIFEXITED(status) && WEXITSTATUS(status) == 0) { /* all ok */
    if (box) {
      return sendmd_final(state, box, out_path, out_len, MD_NEW);
    } else {
      return NULL;
    }
  }
  unlink(out_path);
  return "bad exit";
}


