/*
   Copyright 2005 Laurent Wacrenier

   This file is part of PLL lmtpd.

   PLL lmtpd is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   PLL lmtpd is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with PLL lmtpd; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include "config.h"

static char const rcsid[] UNUSED =
"$Id: sievelex.c,v 1.11 2008/06/30 09:57:48 jcpp Exp $";

#include <sys/types.h>
#define WITH_SIEVE_REGEX

#ifdef WITH_SIEVE_REGEX
#include <regex.h>
#endif

#include <stdarg.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <limits.h>

#include "sieve-token.h"

#include "identifier.c"
#include "string.c"
#include "../rfc/rfc.h"
#include "sieve.h"

#if 0
#include <dmalloc.h>
#endif

#define BACKMAXLEN  3

#define LOCAL static inline

struct sieve_state;

struct string_list {
  int len;
  int allocated;
  char **string;
  struct sieve_state *state;
};


LOCAL void free_strings(char **s) {
  if (s) {
    char **l = s;
    while(*l) {
      free(*l);
      l++;
    }
    free(s);
  }
}

LOCAL int get_string_id(char *str) {
  const struct string *s = sieve_string(str, strlen(str));
  return s ? s->op : 0;
}

void sieve_result_free(struct sieve_result *result) {
  free_strings(result->redirect);
  free_strings(result->fileinto);
  free(result->reject);
}

/* optional require */
/* RFC 3028 */
#define REQUIRE_FILEINTO                 0x0001
#define REQUIRE_REJECT                   0x0002
#define REQUIRE_ENVELOPE                 0x0004
/* RFC 2244 */
#define REQUIRE_COMPARATOR_ASCII_NUMERIC 0x0008
/* RFC 3431 */
#define REQUIRE_RELATIONAL               0x0010
/* RFC 3598 */
#define REQUIRE_SUBADDRESS               0x0020
/* RFC 3894 */
#define REQUIRE_COPY                     0x0040
/* RFC 3685 */
#define REQUIRE_VIRUSTEST                0x0080
#define REQUIRE_SPAMTEST                 0x0100
/* draft-murchison-sieve-regex-07 */
#define REQUIRE_REGEX                    0x0200

#define REQUIRE_ALL                      0xFFFF

struct sieve_state {
  int implicit_keep;
  int linenum;
  int charnum;
  char *reject;
  struct string_list fileinto;
  struct string_list redirect;
  unsigned int require;
  unsigned int allowed_require;
  char *error;
  int memerror;
  int testmode;

  int may_use_require;

  /* IO */
  int backlen;
  int back[BACKMAXLEN];
  //  int (*fgetc)(void *);
  sieve_fgetc_t fgetc;
  void *file;

  struct sieve_methods *methods;

  void *message;
};

#define STRING_BUF 100
#define STRING_LIST_BUFF 10

typedef  struct {
  size_t len;
  size_t alloc;
  char *data;
  struct sieve_state *state;
} STRING;


LOCAL void sieve_error(struct sieve_state *state, char *fmt, ...) {
  if (!state->error[0]) {
    va_list ap;
    va_start(ap, fmt);
    vsnprintf(state->error, SIEVE_STRING_LEN-1, fmt, ap);
    state->error[SIEVE_STRING_LEN-1] = 0;
    va_end(ap);
  }
}

LOCAL int sieve_has_error(struct sieve_state *state) {
  return state->error[0] != 0;
}

LOCAL void sieve_missing_require(struct sieve_state *state, 
				 char *require, char *why, char *type) {
  sieve_error(state, "missing 'require \"%s\"' for '%s' %s",
	      require, why, type);
}

LOCAL void sieve_memory_error(struct sieve_state *state) {
  if (!state->error[0]) {
    sieve_error(state, "Cannot allocate memory");
    state->memerror = 1;
  }
}

LOCAL char **string_list_init(struct sieve_state *st, struct string_list *sl) {
  sl->len = 0;
  sl->allocated = 0;
  sl->string = NULL;
  sl->state = st;
  return NULL;
}

LOCAL void string_list_free(struct string_list *sl) {
  while (sl->len>0) {
    free(sl->string[--sl->len]);
  }
  free(sl->string);
}

LOCAL char **string_list_add(struct string_list *sl, char *s) {
  if (s == NULL) {
    if (sl->len == 0) {
      free(sl->string);
      sl->string = NULL;
      sl->allocated = 0;
      return NULL;
    }
    sl->allocated = sl->len+1; /* cut the string list to NULL */
    sl->string = realloc(sl->string, sizeof(char *) * sl->allocated);
  }
  if (sl->allocated <= sl->len) {
    char **data;
    int allocated = sl->allocated + STRING_LIST_BUFF;
    /* add an extra char* for last NULL later */
    data = realloc(sl->string, (allocated + 1) * sizeof(char *));
    if (data == NULL) { 
      sieve_memory_error(sl->state);
      string_list_free(sl);
      sl->allocated = 0;
      sl->string = NULL;
      return NULL;
    } else {
      sl->string = data;
      sl->allocated = allocated;
    }
  }
  sl->string[sl->len++] = s;
  return sl->string;
}


LOCAL char *string_init (struct sieve_state *st, STRING *s) {
  s->state = st;
  s->len = 0;
  s->data = malloc(STRING_BUF);
  if (s->data == NULL)
    sieve_memory_error(st);
  s->alloc = STRING_BUF;
  return s->data;
}


LOCAL char *string_add(STRING *s, int c) {
  if (c == 0) {
    s->alloc = s->len + 1;
    s->data = realloc(s->data, s->alloc); /* cut the string to \0 */
    /*    fprintf(stderr, "data=%p (%*s)\n", s->data, (int)s->len, s->data); */
  }
  if (s->len >= s->alloc) {
    /* add an extra char for last \0 later */
    char *data = realloc(s->data, s->alloc + STRING_BUF + 1);
    if (data == NULL) {
      sieve_memory_error(s->state);
      free(s->data);
      s->alloc = 0;
      return NULL;
    } else {
      s->alloc += STRING_BUF;
      s->data = data;
    }
  }
  s->data[s->len++] = c;
  return s->data;
}

LOCAL char *string_tolower(STRING *s) {
  int i = 0;
  while(i < s->len) {
    s->data[i] = tolower(s->data[i]);
    i++;
  }
  return s->data;
}

LOCAL void string_free(STRING *s) {
  free(s->data);
}

LOCAL int xfgetc(struct sieve_state *st) {
  if (st->backlen > 0) {
    return st->back[--st->backlen];
  } else {
    int c = st->fgetc(st->file);
    if (st->charnum == 0)
      st->linenum ++;
    if (c == '\n') {
      st->charnum = 0;
    } else {
      st->charnum++;
    }
    return c;
  }
}

LOCAL int xungetc(int c, struct sieve_state *st) {
  if (st->backlen > BACKMAXLEN || c == EOF) {
    return EOF;
  } else {
    st->back[st->backlen++] = c;
    return c;
  }
}

LOCAL int whitespace(struct sieve_state *state) {
  int cont = 1;
  int hasign = 0;
  do {
    int c = xfgetc(state);
    switch(c) {
    case '#':
      while((c = xfgetc(state)) != EOF && c != '\n');
      hasign = 1;
      break;
    case '\n':
    case '\r':
    case ' ':
    case '\t':
      hasign = 1;
      break;
    case '/':
      if ((c = xfgetc(state)) == '*') {
	hasign = 1;
	while((c = xfgetc(state)) != EOF) {
	  if (c == '*') {
	    c = xfgetc(state);
	    if (c == '/') {
	      break;
	    } else {
	      xungetc(c, state);
	    }
	  }
	}
      } else {
	xungetc(c, state);
	xungetc('/', state);
	cont = 0;
      }
      break;
    default:
      xungetc(c, state);
      cont = 0;
      break;
    }
  } while(cont);
  return hasign;
}

LOCAL void read_text(struct sieve_state *state, struct token *t, int doit) {
  int cont = 1;
  int c0;
  int linenum = state->linenum; /* save line counter for errors */
  STRING id;
  do {
    int c = xfgetc(state);
    switch(c) {
    case ' ': 
    case '\r': 
    case '\t':
      break;
    case '#':
      while((c = xfgetc(state)) != EOF && c != '\n');
      cont = 0;
      break;
    case '\n':
      cont = 0;
      break;
    case EOF:
      sieve_error(state, "unexpected EOF in multi line declaration");
      return;
    default:
      sieve_error(state, "unexpected char '%c' in multi line declaration", c);
      return;
    }
  } while(cont);
  
  if (doit)
    string_init(state, &id);
  c0 = '\n';
  do {
    int c = xfgetc(state);
    if (c == EOF) {
      state->linenum = linenum;
      string_free(&id);
      sieve_error(state, "unterminated multi line");
      return;
    }
    if (c == '.' && c0 == '\n') {
      c = xfgetc(state);
      if (c == '\r') 
	c = xfgetc(state);
      if (c == '\n') {
	break;
      }
      if (doit && c != '.') /* skip first dot only */
	string_add(&id, '.');
    }
    if (doit && c != '\r')
      string_add(&id, c);
    c0 = c;
  } while(1);

  t->op = SIEVE_STRING;
  t->value.string = doit ? string_add(&id, 0) : NULL;
}

LOCAL void identifier(struct sieve_state *state, struct token *t, int doit) {
  int c;
  const struct identifier *idx;
  STRING id;
  string_init(state, &id);
  while((c = xfgetc(state)) == '_' || isalnum(c) || c == ':') {
      string_add(&id, c);
  }
  xungetc(c, state);
  string_add(&id, 0);
  string_tolower(&id);
  /*  fprintf(stderr, "identifier %s\n", id.data); */
  idx = sieve_token(id.data, id.len-1);
  if (idx) {
    if (idx->op == SIEVE_TEXT) {
      read_text(state, t, doit);
    } else {
      t->op = idx->op;
      t->value.string = idx->name;
    }
  } else {
    sieve_error(state, "unknown identifier '%s'", id.data);
    t->op = id.data[0] == ':' ? SIEVE_TAG : SIEVE_IDENTIFIER;
  }
  string_free(&id);
}

LOCAL void quoted_string(struct sieve_state *state, struct token *t, int doit) {
  int c;
  STRING id;
  int linenum = state->linenum;
  if (doit)
    string_init(state, &id);
  xfgetc(state); /* pass the '"' */
  while((c = xfgetc(state)) != '"') {
    if (c == '\\') {
      c = xfgetc(state);
    }
    if (c == EOF) {
      state->linenum = linenum;
      sieve_error(state, "unterminated quoted string"); 
      string_free(&id);
      return;
    }
    if (doit && c != '\r')
      string_add(&id, c);
  }
  t->op = SIEVE_STRING;
  t->value.string = doit ? string_add(&id, 0) : NULL;
}

LOCAL void number(struct sieve_state *state, struct token *t, int doit) {
  int c = xfgetc(state);
  unsigned long ret = c - '0';
  while(isdigit(c = xfgetc(state))) {
    ret = ret * 10 + (c - '0');
  }
  switch(c) {
  case 'K':
    ret *= 1024; break;
  case 'M':
    ret *= 1024 * 1024; break;
  case 'G':
    ret *= 1024 * 1024 * 1024; break;
  default:
    xungetc(c, state);
  }
  t->op = SIEVE_NUMBER;
  t->value.number = ret;
}

LOCAL void sieve_lex(struct sieve_state *state, struct token *t, int doit) {
  if (state->error[0]) { /* just in case */
    t->op = SIEVE_EOF;
    return;
  }
  while(1) {
    int c = xfgetc(state);
    switch(c) {
    case EOF:
      t->op = SIEVE_EOF;
      return;
    case ',': 
    case ';': 
    case '(': 
    case ')': 
    case '{': 
    case '}': 
    case '[': 
    case ']': 
      t->op = c;
      t->value.string = "ponct";
      return;
    }
    xungetc(c, state);
    if (isalpha(c) || c == '_' || c == ':') {
      identifier(state, t, doit);
    } else if (isdigit(c)) {
      number(state, t, doit);
    } else if (c == '"') {
      quoted_string(state, t, doit); 
    } else if(whitespace(state)) {
      continue;
      /* ignore */
    } else {
      sieve_error(state, "unexpected character '%c'", c);
      t->op = SIEVE_EOF;
    }
    return;
  }
}

LOCAL void sieve_lex_free(struct token *t) {
  switch(t->op) {
  case SIEVE_STRING:
    free(t->value.string);
    break;
  }
}

LOCAL char **sieve_string_list(struct sieve_state *state, int doit,
				struct token *first) {
  struct token Pt;
  struct string_list Psl;
  if (doit)
    string_list_init(state, &Psl);
  if (!first) {
    sieve_lex(state, &Pt, doit);
    first = &Pt;
  }
  if (first->op == SIEVE_STRING) {
    if (doit) {
      string_list_add(&Psl, first->value.string);
      return string_list_add(&Psl, NULL);
    } else {
      return NULL;
    }
  } else if (first->op == '[') {
    do {
      sieve_lex(state, &Pt, doit);
      if (Pt.op == SIEVE_STRING) {
	if (doit)
	  string_list_add(&Psl, Pt.value.string);
      } else {
	sieve_error(state, "string expected in string list");
	if (doit)
	  string_list_free(&Psl);
	return NULL;
      }
      sieve_lex(state, &Pt, 0);
      if (Pt.op == ',')
	continue;
      if (Pt.op == ']')
	break;
      if (Pt.op != ']') {
	sieve_error(state, "unexpected token building string list");
	if (doit)
	  string_list_free(&Psl);
	break;
      }
    } while(1);
  } else {
    sieve_error(state, "string list expected");
    return NULL;
  }
  return doit ? string_list_add(&Psl, NULL) : NULL;
}

LOCAL int sieve_expect(struct sieve_state *state, int token) {
  struct token Pt;
  sieve_lex(state, &Pt, 0);
  if (Pt.op != token) {
    sieve_error(state, "missing '%c'", token, Pt.op);
    return -1;
  }
  return 0;
}


#define COMPARABLE_OCTET     0
#define COMPARABLE_NUMBER    1
#define COMPARABLE_REGEX     2
#define COMPARABLE_EOL      -1

struct comparable {
  int type;
  union {
    unsigned long number;
    char *octet;
#ifdef WITH_SIEVE_REGEX
    regex_t regex;
#endif
  } value;
  void *alloc;
};

LOCAL void comparable_free(struct comparable *c) {
  switch(c->type) {
  case COMPARABLE_OCTET:
    free(c->alloc);
    break;
#ifdef WITH_SIEVE_REGEX
  case COMPARABLE_REGEX:
    regfree(&c->value.regex);
#endif
  }
}


LOCAL int comparable_numeric(struct sieve_state *state,
			     struct comparable *c) {
  switch(c->type) {
  case COMPARABLE_OCTET:
    {
      char *str = c->value.octet;
      /*      fprintf(stderr, "convert '%s'\n", str); */
      c->value.number = 0;
      if (*str < '0' || *str > '9') {
	c->value.number = ULONG_MAX;
      }
      while(*str >= '0' && *str <= '9') {
	c->value.number = 10*c->value.number + (*str - '0');
	str++;
      }
      free(c->alloc);
      c->type = COMPARABLE_NUMBER;
      /*      fprintf(stderr, "converted %lu\n", c->value.number); */
    }
    break;
  case COMPARABLE_NUMBER:
    return 0;
  }
  return 0;
}

LOCAL int comparable_empty(struct sieve_state *state, struct comparable *s) {
  switch(s->type) {
#ifdef WITH_SIEVE_REGEX
  case COMPARABLE_REGEX:
    regfree(&s->value.regex);
    break;
#endif
  case COMPARABLE_OCTET:
    free(s->alloc);
    break;
  default:
    s->type = COMPARABLE_OCTET;
    break;
  }
  s->value.octet = "";
  s->alloc = NULL;
  return 0;
}

LOCAL int comparable_octet(struct sieve_state *state, struct comparable *s) {
  switch(s->type) {
  case COMPARABLE_OCTET:
    break;
  case COMPARABLE_NUMBER:
    {
      char buf[100]; /* should be enought */
      snprintf(buf, sizeof(buf)-1, "%lu", s->value.number);
      buf[sizeof(buf)-1] = 0;
      s->value.octet = strdup(buf);
      s->alloc = s->value.octet;
      s->type = COMPARABLE_OCTET;
    }
    break;
  }
  return 0;
}

#ifdef WITH_SIEVE_REGEX
LOCAL int comparable_regex(struct sieve_state *state,
			   struct comparable *c) {
  switch(c->type) {
  case COMPARABLE_NUMBER:
    comparable_octet(state, c);
    /* FALLTHROUGH */
  case COMPARABLE_OCTET: {
    char *pattern = c->value.octet;
    int ret = regcomp(&c->value.regex, pattern, REG_EXTENDED|REG_NOSUB);
    if (ret) {
      char errbuf[1000];
      regerror(ret, &c->value.regex, errbuf, sizeof(errbuf)-1);
      sieve_error(state, "compiling regexp \"%.100s\": %s", pattern, errbuf);
      comparable_empty(state, c);
      return -1;
    } else {
      free(c->alloc);
      c->type = COMPARABLE_REGEX;
    }
    break;
  case COMPARABLE_REGEX:
    return 0;
  default: 
    return -1;
  }
  }
  return 0;
}
#endif


LOCAL int comparable_domain(struct sieve_state *state, struct comparable *s) {
  switch(s->type) {
  case COMPARABLE_OCTET:
    {
      char *d = strchr(s->value.octet, '@');
      if (d) 
	s->value.octet = d+1;
      else {
	comparable_empty(state, s);
      }
      return 0;
    }
    /*  case COMPARABLE_NUMBER: */
  }
  comparable_empty(state, s);
  return 0;
}

LOCAL int comparable_localpart(struct sieve_state *state, 
			       struct comparable *s) {
  switch(s->type) {
  case COMPARABLE_OCTET:
    {
      char *d = strchr(s->value.octet, '@');
      if (d) {
	if (s->alloc) {
	  *d = 0;
	} else {
	  char *localpart = malloc(d - s->value.octet + 1);
	  if (localpart == NULL) {
	    sieve_memory_error(state);
	    return comparable_empty(state, s);
	  } 
	  memcpy(localpart, s->value.octet, d - s->value.octet);
	  localpart[ d - s->value.octet ] = 0;
	  s->value.octet = localpart;
	  free(s->alloc);
	  s->alloc = localpart;
	}	  
      }
    }
    break;
  case COMPARABLE_NUMBER:
    break;
  }
  return comparable_empty(state, s);
}

LOCAL int comparable_detail(struct sieve_state *state, struct comparable *s) {
  switch(s->type) {
  case COMPARABLE_OCTET:
    if (state->methods->separator) {
      char *domain = strchr(s->value.octet, '@');
      char *detail = strchr(s->value.octet, state->methods->separator);
      if (detail && detail < domain) {
	if (s->alloc) {
	  *domain = 0;
	  s->value.octet = detail + 1;
	} else {
	  char *alloc = malloc(domain - detail);
	  if (alloc == NULL) {
	    sieve_memory_error(state);
            return comparable_empty(state, s);
	  }
	  memcpy(alloc, detail, domain - detail - 1);
	  alloc[domain - detail - 1] = 0;
	  s->alloc = alloc;
	  s->value.octet = alloc;
	}
	return 0;
      }
    }
  case COMPARABLE_NUMBER:
    break ;
  }
  return comparable_empty(state, s);
}

LOCAL int comparable_user(struct sieve_state *state, struct comparable *s) {
  switch(s->type) {
  case COMPARABLE_OCTET:
    if (state->methods->separator) {
      char *domain = strchr(s->value.octet, '@');
      char *detail = strchr(s->value.octet, state->methods->separator);
      if (detail && detail < domain) {
	if (s->alloc) {
	  *detail = 0;
	} else {
	  char *alloc = malloc(detail - s->value.octet + 1);
	  if (alloc == NULL) {
            sieve_memory_error(state);
            return comparable_empty(state, s);
          }
	  memcpy(alloc, s->value.octet, detail - s->value.octet);
	  alloc[domain - detail] = 0;
	  s->alloc = alloc;
	  s->value.octet = alloc;
	}
	return 0;
      }
    }
    comparable_empty(state, s);
    return 0;
  case COMPARABLE_NUMBER:
    break ;
  }
  return comparable_empty(state, s);
}

LOCAL int comparable_asciiupper(struct sieve_state *state, 
				struct comparable *s) {
  switch(s->type) {
  case COMPARABLE_OCTET:
    {
      char *src;
      if (s->alloc == NULL) {
	size_t len = strlen(s->value.octet)+1;
	s->alloc = malloc(len);
	if (s->alloc == NULL) {
	  sieve_memory_error(state);
	  return 0;
	}
	memcpy(s->alloc, s->value.octet, len);
	s->value.octet = s->alloc;
      }
      src = s->value.octet;
      while(*src) {
	if (*src >= 'a' && *src <= 'z') {
	  *src += ('A' - 'a');
	}
	src++;
      }
    }
    break;
  case COMPARABLE_NUMBER:
    return comparable_octet(state, s);
  }
  return 0;
}


/*
 * convert a list of strings to a list of comparables
 */

LOCAL struct comparable *strings_to_comparables(struct sieve_state *state, 
						char **sl, int alloc) {
  if (sl) {
    char **slc = sl;
    struct comparable *cl, *clc;
    while(*slc) 
      slc++;
    cl = malloc((slc - sl) * sizeof(struct comparable));
    if (cl == NULL) {
      sieve_memory_error(state);
      return NULL;
    }
    clc = cl;
    slc = sl;
    while(*slc) {
      clc->type = COMPARABLE_OCTET;
      clc->value.octet = *slc;
      clc->alloc = alloc ? *slc : NULL;
      clc++; slc ++;
    }
    return cl;
  } else {
    return NULL;
  }
}


LOCAL int test_relational(struct sieve_state *state,
			  struct comparable *c1, struct comparable *c2) {
  if (c1->type != c2->type)
    return 0;
  switch(c1->type) {
  case COMPARABLE_OCTET:
    /*    fprintf(stderr, "compare \"%s\" with \"%s\"", c1->value.octet, c2->value.octet);  */
    return strcmp(c2->value.octet, c1->value.octet);
  case COMPARABLE_NUMBER:
    /*    fprintf(stderr, "compare %lu with %lu\n", c1->value.number,  c2->value.number); */
    if (c2->value.number > c1->value.number) {
      return 1;
    } else if (c2->value.number < c1->value.number) {
      return -1;
    } else {
      return 0;
    }
  }
  return 0;
}

LOCAL int test_relational_gt(struct sieve_state *state,
			     struct comparable *c1, struct comparable *c2) {
  return test_relational(state, c1, c2) > 0;
}
LOCAL int test_relational_ge(struct sieve_state *state,
			     struct comparable *c1, struct comparable *c2) {
  return test_relational(state, c1, c2) >= 0;
}
LOCAL int test_relational_lt(struct sieve_state *state,
			     struct comparable *c1, struct comparable *c2) {
  return test_relational(state, c1, c2) < 0;
}
LOCAL int test_relational_le(struct sieve_state *state,
			     struct comparable *c1, struct comparable *c2) {
  return test_relational(state, c1, c2) <= 0;
}
LOCAL int test_relational_eq(struct sieve_state *state,
			     struct comparable *c1, struct comparable *c2) {
  return test_relational(state, c1, c2) == 0;
}
LOCAL int test_relational_ne(struct sieve_state *state,
			     struct comparable *c1, struct comparable *c2) {
  return test_relational(state, c1, c2) != 0;
}

LOCAL int test_is(struct sieve_state *state,
		  struct comparable *c1, struct comparable *c2) {
  if (c1->type != c2->type) 
    return 0;
  switch (c1->type) {
  case COMPARABLE_OCTET:
    /*    fprintf(stderr, "compare(%s, %s) = %d\n", c1->value.octet, c2->value.octet, strcmp(c1->value.octet, c2->value.octet) == 0); */
    return strcmp(c1->value.octet, c2->value.octet) == 0;
  case COMPARABLE_NUMBER:
    return c1->value.number == c2->value.number;
  default: 
    return 0;
  }
}

#ifdef WITH_SIEVE_REGEX
LOCAL int test_regex(struct sieve_state *state,
		      struct comparable *c1, struct comparable *c2) {
  /* yet another exeption, this should be deported elsewhere */
  if (c1->type != COMPARABLE_REGEX){
    if (comparable_regex(state, c1) != 0)
      return 0;
  }
  if (c2->type != COMPARABLE_OCTET) {
    comparable_octet(state, c2);
  }
  return regexec(&c1->value.regex, c2->value.octet, 0, NULL, 0) == 0;
}
#endif

LOCAL int test_contains(struct sieve_state *state,
			struct comparable *c1, struct comparable *c2) {
  if (c1->type != COMPARABLE_OCTET || c2->type != COMPARABLE_OCTET) {
    return 0;
  }
  /*  fprintf(stderr, "contains %s, %s\n", c1->value.octet, c2->value.octet); */
  return strstr(c2->value.octet, c1->value.octet) != 0;
}



LOCAL int sieve_matches(char *s1, char *s2) {
  unsigned char *string = (unsigned char*)s1;
  unsigned char *pattern = (unsigned char *)s2;
  while(*pattern) {
    switch(*pattern) {
    case '?': 
      {
	int len;
	pattern ++;
	if (*string == 0)
	  return 0;
	if (*string < 0x80)      /* 1000 0000 */
	  len = 0;
	else if (*string < 0xC0) /* 1100 0000 */
	  return 0;
	else if (*string < 0xE0) /* 1110 0000 */
	  len = 1;
	else if (*string < 0xF0) /* 1111 0000 */
	  len = 2;
	else if (*string < 0xF8) /* 1111 1000 */
	  len = 3;
	else if (*string < 0xFC) /* 1111 1100 */
	  len = 4;
	else if (*string < 0xFE) /* 1111 1110 */
	  len = 5;
	else 
	  return 0;
	string++;
	while(len-->0) {
	  if (*string == 0 || *string < 0x80 || *string >= 0xC0)
	    return 0;
	  string++;
	}
	/*
	if (strlen(string) < len)
	  return 0;
	string += len;
	*/
      }
      break;
    case '*': 
      while(*pattern == '*')
	pattern++;
      if (*pattern == 0)
	return 1;
      while(*string) {
	if (sieve_matches((char*)string, (char*)pattern))
	  return 1;
	string++;
      }
      return 0;
    case '\\':
      pattern ++;
      if (*string++ != *pattern++)
	return 0;
      break;
    case '\0':
      return *string == 0;
    default:
      if (*string++ != *pattern++)
	return 0;
    }
  }
  return 1;
}

LOCAL int test_matches(struct sieve_state *state,
		       struct comparable *c1, struct comparable *c2) {
  if (c1->type != COMPARABLE_OCTET || c2->type != COMPARABLE_OCTET) 
    return 0;
  return sieve_matches(c1->value.octet, c2->value.octet);
}


/* options */
#define CAMS_COMPARATOR    0x01
#define CAMS_MATCH_TYPE    0x02
#define CAMS_ADDRESS_PART  0x04

/* terminators */
#define CAMS_STRING_LIST1  0x08
#define CAMS_STRING_LIST2  0x10
#define CAMS_STRING        0x20

struct cams {
  int (*match_type)(struct sieve_state *state, 
		    struct comparable *, struct comparable *);
  int (*address_part)(struct sieve_state *state, struct comparable *);
  int (*comparator)(struct sieve_state *state, struct comparable *);
  int count;
  char **strings1;
  char **strings2;
  char *string;
};

#define CHECK_FOR_ARG(x, type) \
   if (what & x) { \
    what &= ~x;    \
   } else {        \
    sieve_error(state, "unexpected " type); \
    return sieve_cams_error(c); \
   }

LOCAL int sieve_cams_error(struct cams *c) {
  free(c->string);
  free_strings(c->strings1);
  free_strings(c->strings2);
  return -1;
}

LOCAL int sieve_cams(struct sieve_state *state, int doit,
		      unsigned int what, struct cams *c) {
  c->comparator = comparable_asciiupper;
  c->match_type = test_is;
  c->address_part = comparable_octet;
  c->count = 0;
  c->strings1 = NULL;
  c->strings2 = NULL;
  c->string = NULL;
  /*  int relational = 0; */
  while(what) {
    struct token Pt;
    sieve_lex(state, &Pt, doit);
    switch(Pt.op) {
    case SIEVE_COMPARATOR:
      CHECK_FOR_ARG(CAMS_COMPARATOR, ":comparator");
      sieve_lex(state, &Pt, doit);
      if (Pt.op == SIEVE_STRING) {
	if  (doit) {
	  int sid = get_string_id(Pt.value.string);
	  sieve_lex_free(&Pt);
	  switch(sid) {
	  case SIEVE_STRING_I_ASCII_CASEMAP: 
	    c->comparator = comparable_asciiupper; break;
	  case SIEVE_STRING_I_ASCII_NUMERIC:
	    if (state->require & REQUIRE_COMPARATOR_ASCII_NUMERIC)
	      c->comparator = comparable_numeric;
	    else {
	      sieve_missing_require(state, "comparator-i;ascii-numeric",
				    "i;ascii-numeric", "comparator");
	      return sieve_cams_error(c);
	    }
	    break;
	  case SIEVE_STRING_I_OCTET:
	    c->comparator = comparable_octet; break;
	  default:
	    sieve_error(state, "unknown comparator");
	    return -1;
	  }
	}
      } else {
	sieve_error(state, "no comparator name");
	return sieve_cams_error(c);
      }
      break;
    case SIEVE_COUNT:
      c->count = 1;
    case SIEVE_VALUE:
      CHECK_FOR_ARG(CAMS_MATCH_TYPE, "match type");
      if (!(state->require & REQUIRE_RELATIONAL)) {
	sieve_missing_require(state, "relational",
			      Pt.value.string, "operator");
	return sieve_cams_error(c);
      }
      sieve_lex(state, &Pt, doit);
      if (Pt.op != SIEVE_STRING) {
	sieve_lex_free(&Pt);
	sieve_error(state, "relational operator need a string argument");
	return -1;
      } else  {
	if (doit) {
	  int sid = get_string_id(Pt.value.string);
	  sieve_lex_free(&Pt);
	  switch(sid) {
	  case SIEVE_STRING_GT: c->match_type = test_relational_gt; break;
	  case SIEVE_STRING_GE: c->match_type = test_relational_ge; break;
	  case SIEVE_STRING_LT: c->match_type = test_relational_lt; break;
	  case SIEVE_STRING_LE: c->match_type = test_relational_le; break;
	  case SIEVE_STRING_EQ: c->match_type = test_relational_eq; break;
	  case SIEVE_STRING_NE: c->match_type = test_relational_ne; break;
	  default:
	    sieve_error(state, "unknown relational operator");
	    return -1;
	  }
	}
      }
      break;
#ifdef WITH_SIEVE_REGEX
    case SIEVE_REGEX:
      CHECK_FOR_ARG(CAMS_MATCH_TYPE, "match type");
      c->match_type = test_regex;
      break;
#endif
    case SIEVE_IS:
      CHECK_FOR_ARG(CAMS_MATCH_TYPE, "match type");
      c->match_type = test_is;
      break;
    case SIEVE_CONTAINS:
      CHECK_FOR_ARG(CAMS_MATCH_TYPE, "match type");
      c->match_type = test_contains;
      break;
    case SIEVE_MATCHES:
      CHECK_FOR_ARG(CAMS_MATCH_TYPE, "match type");
      c->match_type = test_matches;
      break;
    case SIEVE_DOMAIN:
      CHECK_FOR_ARG(CAMS_ADDRESS_PART, "address part");
      c->address_part = comparable_domain;
      break;
    case SIEVE_ALL:
      CHECK_FOR_ARG(CAMS_ADDRESS_PART, "address part");
      c->address_part = comparable_octet;
      break;
    case SIEVE_LOCALPART:
      CHECK_FOR_ARG(CAMS_ADDRESS_PART, "address part");
      c->address_part = comparable_localpart;
      break;
    case SIEVE_USER:
      if (state->require & REQUIRE_SUBADDRESS) {
	CHECK_FOR_ARG(CAMS_ADDRESS_PART, "address part");
	c->address_part = comparable_user;
      } else {
	sieve_missing_require(state, "subaddress", ":user", "address part");
	return -1;
      }
      break;
    case SIEVE_DETAIL:
      if (state->require & REQUIRE_SUBADDRESS) {
	CHECK_FOR_ARG(CAMS_ADDRESS_PART, "address part");
	c->address_part = comparable_detail;
      } else {
	sieve_missing_require(state, "subaddress", ":detail", "address part");
	return sieve_cams_error(c);
      }
      break;
    case SIEVE_STRING:
      if (what & CAMS_STRING) {
	c->string = Pt.value.string;
	what = 0;
	break;
      }
      /* FALLTHROUGH */
    case '[':
      if (what & (CAMS_STRING_LIST1 | CAMS_STRING_LIST2)) {
	char **sl = sieve_string_list(state, doit, &Pt);
	if (sl == NULL && sieve_has_error(state)) {
	  sieve_error(state, "invalid string list");
	  return sieve_cams_error(c);
	}
	if (what & CAMS_STRING_LIST1) {
	  what &= CAMS_STRING_LIST2;
	  c->strings1 = sl;
	} else {
	  what = 0;
	  c->strings2 = sl;
	}
      }
      break;
    case SIEVE_EOF:
      sieve_error(state, "unexpected EOF");
      return sieve_cams_error(c);
    default:
      sieve_error(state, "unexpected token");
      return sieve_cams_error(c);
      }
  } 
  return 0;
}

struct sieve_get_from_mail {
  char **headers;
  void *state;
  char *list;
  int (*counter)(struct sieve_state *, struct sieve_get_from_mail *,
		 struct comparable *);
  void (*state_free)(struct sieve_state *, void *state);
};

LOCAL void sieve_header_state_free(struct sieve_state *st, void *state) {
  if (state && st->methods->header_state_free) {
    st->methods->header_state_free(st->message, state);
  }
}

LOCAL int sieve_get_count(struct sieve_state *state,
			  struct sieve_get_from_mail *context,
			  struct comparable *c) {
  if (context->counter) {
    int count;
    unsigned long total = 0;
    while((count = context->counter(state, context, NULL))>= 0)
      total += count;
    if (c) {
      c->type = COMPARABLE_NUMBER;
      c->value.number = total;
      c->alloc = NULL;
    }
    context->counter = NULL;
    return 1;
  } else {
    return -1;
  }
}


LOCAL int sieve_get_envelope(struct sieve_state *state, 
			     struct sieve_get_from_mail *context,
			     struct comparable *c) {
  char *result = NULL;
  char **headers = context->headers;
  int ret = -1;
  while(*headers && ret<0) {
    if (strcasecmp(*headers, "from") == 0) {
      if (state->methods->from) 
	result = state->methods->from(state->message) ;
      if (!result)
	result = "";
      ret = 1;
    } else if (strcasecmp(*headers, "to") == 0) {
      if (state->methods->to)
	result = state->methods->to(state->message);
      if (!result)
	result = "";
      ret = (*result != 0);
    }
    headers ++;
  }
  context->headers = headers;
  if (c) {
    c->type = COMPARABLE_OCTET;
    c->value.octet = result;
    c->alloc = NULL;
  }
  return ret;
}

LOCAL int sieve_get_header(struct sieve_state *state,
			   struct sieve_get_from_mail *context,
			   struct comparable *c) {
  if (state->methods->header) {
    char *result = NULL;
    context->state_free = sieve_header_state_free;
    while(!result && *context->headers) {
      /*      fprintf(stderr, "headers=%s state=%p\n", *context->headers, context->state); */
      result = state->methods->header(state->message,
				      *context->headers,
				      1, &context->state);
      if (result == NULL)
	context->headers ++;
    }
    if (result) {
      if (c) {
	c->type = COMPARABLE_OCTET;
	c->value.octet = result;
	c->alloc = NULL;
      }
      return 1;
    } else {
      return -1;
    }
  } else {
    return -1;
  }
}

LOCAL int sieve_get_address(struct sieve_state *state,
			    struct sieve_get_from_mail *context,
			    struct comparable *c) {
  char *result = NULL;
  if (state->methods->header == NULL)
    return -1;

  context->state_free = sieve_header_state_free;

  while(1) {
    while(*context->headers && context->list == NULL) {
      context->list = state->methods->header(state->message,
					     *context->headers,
					     0, &context->state);
      /*      fprintf(stderr, "list='%s', ch=%s\n", context->list, *context->headers);  */
      if (context->list == NULL) {
	context->headers ++;
      }
    }
    if (context->list) {
      result = rfc822_getaddr(&context->list);
      if (*context->list == 0) {
	context->list = NULL;
	context->headers ++;
      }
      /*      fprintf(stderr, "address = <%s>, list='%s', ch=%s\n", result, context->list, *context->headers);  */
      if (result)
	break;
    }
    if (*context->headers == NULL && context->list == NULL)
      return -1;
  }
  if (c) {
    c->type = COMPARABLE_OCTET;
    c->value.octet = result;
    c->alloc = result;
  } else {
    free(result);
  }

  return 1;
}

LOCAL int sieve_test_simpletest(struct sieve_state *state, int doit,
				unsigned (*handle)(void *)) {
  struct cams c;
  if (sieve_cams(state, doit,
		 CAMS_COMPARATOR|CAMS_MATCH_TYPE|CAMS_STRING, &c) == -1)
    return -1;
  if (doit) {
    struct comparable res;
    struct comparable arg;
    int ret;

    res.type = COMPARABLE_NUMBER;
    res.value.number = handle ? handle(state->message) : 0;
    /*    fprintf(stderr, "simpletest: n=%lu\n", res.value.number); */
    arg.type = COMPARABLE_OCTET;
    arg.value.octet = c.string;
    arg.alloc = c.string;

    c.comparator(state, &res);
    c.comparator(state, &arg);
    ret = c.match_type(state, &arg, &res);
    comparable_free(&res);
    comparable_free(&arg);
    return ret;
  } else {
    return 0;
  }
}

LOCAL char **string_list_tolower(char **sl, int doit) {
  char **slr = sl;
  if (doit && sl) {
    while(*sl) {
      char *s = *sl;
      while(*s) {
	*s = tolower(*s);
	s++;
      }
      sl++;
    }
  }
  return slr;
}

LOCAL int sieve_test_multi(struct sieve_state *state, int doit,
			    unsigned int args,
			    int (*get_data)(struct sieve_state *,
					    struct sieve_get_from_mail *,
					    struct comparable *)
			    ) {
  struct cams c;
  int ret = 0;
  if (sieve_cams(state, doit, args, &c) == -1)
    return -1;
  if (doit) {
    struct comparable *complist = strings_to_comparables(state, c.strings2, 1);
    struct comparable *comp;
    int list2_parsed = 0;
    struct sieve_get_from_mail context = {0,};
    struct comparable s;
    char **str;
    int count = 0;

    context.headers = string_list_tolower(c.strings1, doit);
    
    if (c.count) { /* yet another syntax exeption */
      context.counter = get_data;
      get_data = sieve_get_count;
    }


    while(!ret && (count = get_data(state, &context, &s))>=0) {
      struct comparable *comp = complist;
      str = c.strings2;

      if (args & CAMS_ADDRESS_PART) c.address_part(state, &s);
      c.comparator(state, &s);

      while(*str) {
	if (!list2_parsed)
	  c.comparator(state, comp);
	/*		fprintf(stderr, "compare type %d %d\n", comp->type, s.type);  */
	if (c.match_type(state, comp, &s)) {
	  ret = 1;
	  break;
	}
	str++;
	comp++;
      }
      list2_parsed = 1;
      comparable_free(&s);
    }
    if (context.state_free) {
      context.state_free(state, context.state);
    }
    str = c.strings2;
    comp = complist;
    
    while(*str) {
      comparable_free(comp);
      str++;
      comp++;
    }
    free(complist);
  }
  free(c.strings2);
  free_strings(c.strings1);
  return ret;
}

LOCAL int sieve_test_envelope(struct sieve_state *state, int doit) {
  return sieve_test_multi(state, doit, 
			  CAMS_COMPARATOR|CAMS_ADDRESS_PART|CAMS_MATCH_TYPE|
			  CAMS_STRING_LIST1|CAMS_STRING_LIST2,
			  sieve_get_envelope);
}

LOCAL int sieve_test_address(struct sieve_state *state, int doit) {
  return sieve_test_multi(state, doit, 
			  CAMS_COMPARATOR|CAMS_ADDRESS_PART|CAMS_MATCH_TYPE|
			  CAMS_STRING_LIST1|CAMS_STRING_LIST2,
			  sieve_get_address);
}

LOCAL int sieve_test_header(struct sieve_state *state, int doit) {
  return sieve_test_multi(state, doit, 
			  CAMS_COMPARATOR|CAMS_MATCH_TYPE|
			  CAMS_STRING_LIST1|CAMS_STRING_LIST2,
			  sieve_get_header);
}

LOCAL int sieve_test_exists(struct sieve_state *state, int doit) {
  char **sl = sieve_string_list(state, doit, NULL);
  int ret = 1;
  if (sl == NULL && sieve_has_error(state)) {
    sieve_error(state, "'exists' test expect a list of headers");
    return 0;
  }
  if (doit) {
    char **slc = sl;
    if (sl == NULL)
      return 0;
    while(*slc) {
      if (ret) {
	if (state->methods->header(state->message, *slc, 0, NULL) == NULL)
	  ret = 0;
      }
      free(*slc);
      slc++;
    }
    free(sl);
    return ret;
  }
  return 0;
}

LOCAL int sieve_test_size(struct sieve_state *state, int doit) {
  struct token Porder;
  struct token Plimit;
  size_t size = 0;
  sieve_lex(state, &Porder, 0);
  if (Porder.op != SIEVE_UNDER && Porder.op != SIEVE_OVER) {
    sieve_error(state, "size test expect :over or :under");
    return 0;
  }
  sieve_lex(state, &Plimit, 0);
  if (Plimit.op != SIEVE_NUMBER) {
    sieve_error(state, "size test needs a numerical limit");
    return 0;
  }
  if (state->methods->size)
    size = state->methods->size(state->message);
  if (Porder.op == SIEVE_UNDER) {
    return size < Plimit.value.number;
  } else {
    return size > Plimit.value.number;
  }
}

LOCAL int sieve_require(struct sieve_state *state) {
  char **sl = sieve_string_list(state, 1, NULL);
  char **slc = sl;
  if (sl == NULL) {
    sieve_error(state, "require need a string or string list argument");
    return -1;
  }
  while(*slc) {
    int sid = get_string_id(*slc);
    unsigned int add_require;
    switch(sid) {
    case SIEVE_STRING_COMPARATOR_I_ASCII_CASEMAP:
    case SIEVE_STRING_COMPARATOR_I_OCTET:
      /* always included */
      add_require = 0;
      break;
    case SIEVE_STRING_FILEINTO: add_require = REQUIRE_FILEINTO; break;
    case SIEVE_STRING_ENVELOPE: add_require = REQUIRE_ENVELOPE; break;
    case SIEVE_STRING_REJECT: add_require = REQUIRE_REJECT; break;
    case SIEVE_STRING_COMPARATOR_I_ASCII_NUMERIC:
      add_require = REQUIRE_COMPARATOR_ASCII_NUMERIC; break;
    case SIEVE_STRING_RELATIONAL: add_require = REQUIRE_RELATIONAL; break;
    case SIEVE_STRING_COPY: add_require = REQUIRE_COPY; break;
    case SIEVE_STRING_SPAMTEST: add_require = REQUIRE_SPAMTEST; break;
    case SIEVE_STRING_VIRUSTEST: add_require = REQUIRE_VIRUSTEST; break;
    case SIEVE_STRING_SUBADDRESS: add_require = REQUIRE_SUBADDRESS; break;
#ifdef WITH_SIEVE_REGEX
    case SIEVE_STRING_REGEX: add_require = REQUIRE_REGEX; break;
#endif
    default:
      sieve_error(state, "unknown requirement \"%s\"", *slc);
      free_strings(sl);
      return -1;
    }
    if (add_require) {
      if (add_require & state->allowed_require) {
	state->require |= add_require;
      } else {
	sieve_error(state, "unsupported requirement \"%s\"", *slc);
	free_strings(sl);
	return -1;
      }
    }
    slc++;
  }
  free_strings(sl);
  return sieve_expect(state, ';');
}

LOCAL int sieve_discard(struct sieve_state *state, int doit) {
  if (sieve_expect(state, ';') == -1)
    return -1;
  if (doit)
    state->implicit_keep = 0;
  return 0;
}

LOCAL int sieve_stop(struct sieve_state *state, int doit) {
  if (sieve_expect(state, ';') == -1)
    return -1;
  if (doit)
    return -1;
  else
    return 0;
}

LOCAL int sieve_fileinto(struct sieve_state *state, int doit) {
  struct token Pt;
  int copy = 0;
  if (!(state->require & REQUIRE_FILEINTO)) {
    sieve_missing_require(state, "fileinto", "fileinto", "command");
    return -1;
  }
  sieve_lex(state, &Pt, 1);
  if (Pt.op == SIEVE_COPY) {
    if (!(state->require & REQUIRE_COPY)) {
      sieve_missing_require(state, "copy", ":copy", "modifier");
      return -1;
    }
    copy = 1;
    sieve_lex(state, &Pt, 1);
  }
  if (Pt.op != SIEVE_STRING) {
    sieve_lex_free(&Pt);
    sieve_error(state, "fileinto expect a folder name");
    return -1;
  }
  if (sieve_expect(state, ';') == -1) {
    sieve_lex_free(&Pt);
    return -1;
  }
  if (state->methods->valid_folder(state->message, Pt.value.string) == 0) {
    sieve_error(state, "invalid folder name '%s'", Pt.value.string);
    sieve_lex_free(&Pt);
    return -1;
  }
  if (doit && !state->testmode) {
    string_list_add(&state->fileinto, Pt.value.string);
    if (!copy)
      state->implicit_keep = 0;
  } else {
    sieve_lex_free(&Pt);
    return 0;
  }
  return 0;
}

LOCAL int sieve_redirect(struct sieve_state *state, int doit) {
  struct token Pt;
  int copy = 0;
  sieve_lex(state, &Pt, doit);
  if (Pt.op == SIEVE_COPY) {
    if (!(state->require & REQUIRE_COPY)) {
      sieve_missing_require(state, "copy", ":copy", "modifier");
      return -1;
    }
    copy = 1;
    sieve_lex(state, &Pt, doit);
  }
  if (Pt.op != SIEVE_STRING) {
    sieve_lex_free(&Pt);
    sieve_error(state, "redirect expect an address");
    return -1;
  }
  if (sieve_expect(state, ';') == -1)
    return -1;
  if (doit) {
    char *address = Pt.value.string;
    char *a = rfc822_getaddr(&address);
    if (a == NULL) {
      sieve_error(state, "invalid address \"%s\"", Pt.value.string);
      sieve_lex_free(&Pt);
      return -1;
    }
    if (state->testmode) {
      free(a);
    } else {
      string_list_add(&state->redirect, a);
      if (!copy)
	state->implicit_keep = 0;
    }
    sieve_lex_free(&Pt);
  }
  return 0;
}

LOCAL int sieve_keep(struct sieve_state *state, int doit) {
  if (sieve_expect(state, ';') == -1)
    return -1;
  if (doit && !state->testmode) {
    string_list_add(&state->fileinto, strdup("INBOX"));
    state->implicit_keep = 0;
  }
  return 0;
}

LOCAL int sieve_reject(struct sieve_state *state, int doit) {
  struct token Pt;
  if (!(state->require & REQUIRE_REJECT)) {
    sieve_missing_require(state, "reject", "reject", "command");
    return -1;
  }
  sieve_lex(state, &Pt, doit);
  if (Pt.op != SIEVE_STRING) {
    sieve_lex_free(&Pt);
    sieve_error(state, "reject expect a reason");
    return -1;
  }
  if (sieve_expect(state, ';') == -1)
    return -1;
  if (doit && !state->testmode) {
    if (state->reject)
      free(state->reject);
    state->reject = Pt.value.string;
    state->implicit_keep = 0;
  } else {
    sieve_lex_free(&Pt);
  }
  return 0;
}

LOCAL int sieve_test(struct sieve_state *state, int doit);

LOCAL int sieve_handle_test_list(struct sieve_state *state, int doit, int type) {
  struct token Pt;
  int value;
  sieve_lex(state, &Pt, 0);
  value = (type == SIEVE_ALLOF);
  if (Pt.op == '(') {
    do {
      int test = sieve_test(state, doit);
      switch(type) {
      case SIEVE_ANYOF:
	if (test) {
	  value = 1;
	  doit = 0;
	}
	break;
      case SIEVE_ALLOF:
	if (!test) {
	  value = 0;
	  doit = 0;
	}
	break;
      }
      sieve_lex(state, &Pt, 0);
      if (Pt.op == ',')
	continue;
      if (Pt.op == ')')
	break;
      sieve_error(state, "comma expected, recieved %d", Pt.op);
      return -1;
    } while(1);
    return value;
  }
  sieve_error(state, "test list expected, recieved %d", Pt.op);
  return -1;
}

LOCAL int sieve_test(struct sieve_state *state, int doit) {
  struct token Pt;
  if (state->testmode) doit = 1; /* check if the script work */
  sieve_lex(state, &Pt, 0);
  switch(Pt.op) {
  case SIEVE_TRUE:
    return 1;
  case SIEVE_FALSE:
    return 0;
  case SIEVE_NOT:
    return !sieve_test(state, doit);
  case SIEVE_ANYOF:
  case SIEVE_ALLOF:
    return sieve_handle_test_list(state, doit, Pt.op);
  case SIEVE_ENVELOPE:
    if (!(state->require & REQUIRE_ENVELOPE)) {
      sieve_missing_require(state, "envelope", "envelope", "test");
      return -1;
    }
    return sieve_test_envelope(state, doit);
  case SIEVE_ADDRESS:
    return sieve_test_address(state, doit);
  case SIEVE_HEADER:
    return sieve_test_header(state, doit);
  case SIEVE_EXISTS:
    return sieve_test_exists(state, doit);
  case SIEVE_SIZE:
    return sieve_test_size(state, doit);
  case SIEVE_SPAMTEST: 
    if (!(state->require & REQUIRE_SPAMTEST)) {
      sieve_missing_require(state, "spamtest", "spamtest", "test");
      return -1;
    }
    return sieve_test_simpletest(state, doit, state->methods->spam);
  case SIEVE_VIRUSTEST: 
    if (!(state->require & REQUIRE_VIRUSTEST)) {
      sieve_missing_require(state, "virustest", "virustest", "test");
      return -1;
    }
    return sieve_test_simpletest(state, doit, state->methods->virus);
  default: 
    sieve_error(state, "unkonwn test %d '%s'", Pt.op, Pt.value.string);
    return -1;
  }  
}

LOCAL int sieve_commands(struct sieve_state *state, int doit, int end);

LOCAL int sieve_block(struct sieve_state *state, int doit) {
  if (sieve_expect(state, '{') == -1)
     return -1;
  if (state->testmode) doit = 1; /* check if the script work */
  return sieve_commands(state, doit, '}');
}

LOCAL int sieve_commands(struct sieve_state *state, int doit, int end_token) {
  struct token Pt;
  int do_else = -1;
  /*
   *  do_else == 0    pass over the next else/elsif
   *  do_else == 1    try the next else/elsif condition
   *  do_else == -1   unexpected else/elsif
   */
  do {
    sieve_lex(state, &Pt, doit);
    if (Pt.op == end_token) {
      return 0;
    }
    if (Pt.op != SIEVE_REQUIRE)
      state->may_use_require = 0;

    switch(Pt.op) {
    case SIEVE_IF:
      {
	int test = sieve_test(state, doit);
	if (test == -1 || sieve_block(state, test && doit) == -1)
	    return -1;
	do_else = doit && !test;
      }
      continue;
    case SIEVE_ELSIF: 
      if (do_else >= 0) {
	int test = sieve_test(state, do_else);
	if (test == -1 || sieve_block(state, test && do_else) == -1)
	  return -1;
	do_else = do_else && !test;
      } else {
	sieve_error(state, "unexpected elsif");
	return -1;
      }
      continue;
    case SIEVE_ELSE:
      if (do_else >= 0) {
	if (sieve_block(state, do_else) == -1)
	  return -1;
	do_else =-1;
      } else {
	sieve_error(state, "unexpected else");
	return -1;
      }
      continue;
    }
    do_else = -1;
    switch(Pt.op) {
    case SIEVE_REQUIRE:
      if (!state->may_use_require) {
	sieve_error(state, "'require' not in filter header");
	return -1;
      }
      if (sieve_require(state) == 0)
	continue;
      else 
	return -1;
    case SIEVE_STOP:
      if (sieve_stop(state, state->testmode ? 0 : doit) == 0) 
	continue;
      else
	return -1;
    case SIEVE_REJECT:
      if (sieve_reject(state, doit) == 0)
	continue;
      else 
	break;
    case SIEVE_FILEINTO:
      if (sieve_fileinto(state, doit) == 0)
	continue;
      else 
	return -1;
    case SIEVE_REDIRECT:
      if (sieve_redirect(state, doit) == 0)
	continue;
      else 
	return -1;
    case SIEVE_KEEP:
      if (sieve_keep(state, doit) == 0)
	continue;
      else 
	return -1;
    case SIEVE_DISCARD:
      if (sieve_discard(state, doit) == 0)
	continue;
      else 
	return -1;
    case SIEVE_EOF:
      if (end_token != SIEVE_EOF)
	sieve_error(state, "unexpected EOF");
      return -1;
    default:
      sieve_error(state, "unexpected token");
      return -1;
    }
  } while(1);
  return -1;
}

LOCAL unsigned int sieve_allowed_extensions(struct sieve_methods *methods) {
  unsigned int allowed_require = REQUIRE_ALL;
  if (methods->from == NULL && methods->to == NULL)
    allowed_require &= ~REQUIRE_ENVELOPE;
  if (methods->valid_folder == NULL) 
    allowed_require &= ~REQUIRE_FILEINTO;
  if (methods->spam == NULL)
    allowed_require &= ~REQUIRE_SPAMTEST;
  if (methods->virus == NULL)
    allowed_require &= ~REQUIRE_VIRUSTEST;
  if (methods->separator == '\0') 
    allowed_require &= ~REQUIRE_SUBADDRESS;
  return allowed_require;
}

LOCAL int sieve_run(sieve_fgetc_t sieve_fgetc, void *file,
		    struct sieve_methods *methods, void *message,
		    struct sieve_result *result,
		    int mode) {
  struct sieve_state Pstate = {0,};
  Pstate.fgetc = sieve_fgetc;
  Pstate.file = file;
  Pstate.message = message;
  Pstate.methods = methods;
  Pstate.implicit_keep = 1;
  Pstate.may_use_require = 1;
  Pstate.error = result->error;
  Pstate.error[0] = 0;
  Pstate.testmode = mode == 1;
  Pstate.allowed_require = sieve_allowed_extensions(methods);

  sieve_commands(&Pstate, 1, SIEVE_EOF);

  if (result) {
    result->implicit_keep = Pstate.implicit_keep;
    if (Pstate.fileinto.string)
      string_list_add(&Pstate.fileinto, NULL);
    if (Pstate.redirect.string)
      string_list_add(&Pstate.redirect, NULL);
    result->fileinto = Pstate.fileinto.string;
    result->redirect = Pstate.redirect.string;
    result->reject = Pstate.reject;
    /*    result->error = Pstate.error; */
    result->lastline = Pstate.linenum;
  }
  if (Pstate.error[0]) {
    if (Pstate.memerror)
      return -1;
    else
      return 1;
  }
  return 0;
}

int sieve(sieve_fgetc_t sieve_fgetc, void *file,
	  struct sieve_methods *methods, void *message,
	  struct sieve_result *result) {
  return sieve_run(sieve_fgetc, file, methods, message, result, 0);
}

LOCAL char *sieve_check_header(void *message, const char *header, 
			       int decoded, void **state) {
  if (state == NULL)
    return NULL;
  if (*state) {
    return NULL;
  } else {
    *state = "";
    return "";
  }
}

LOCAL unsigned sieve_check_unsigned(void *message) {
  return 0;
}

int sieve_check(sieve_fgetc_t sieve_fgetc, void *file,
		struct sieve_methods *methods, void *message,
		struct sieve_result *result) {
  struct sieve_methods Pmethods;
  memcpy(&Pmethods, methods, sizeof(struct sieve_methods));
  Pmethods.header = sieve_check_header;
  if (methods->spam)
    Pmethods.spam = sieve_check_unsigned;
  if (methods->virus)
    Pmethods.spam = sieve_check_unsigned;
  return sieve_run(sieve_fgetc, file, &Pmethods, message, result, 1);
}

char **sieve_extensions(struct sieve_methods *methods) {
  unsigned int allowed_require = sieve_allowed_extensions(methods);
  int n = 1;
  unsigned int xrequire = 1;
  char **capa;
  while(xrequire) {
    n+= xrequire & allowed_require ;
    xrequire <<=1;
  }

  capa = malloc(n * sizeof(char *));

  n = 0; /* reset the counter */
  if (capa == NULL)
    return NULL;
  if (allowed_require & REQUIRE_FILEINTO)
    capa[n++] = "FILEINTO";
  if (allowed_require & REQUIRE_REJECT)
    capa[n++] = "REJECT";
  if (allowed_require & REQUIRE_ENVELOPE)
    capa[n++] = "ENVELOPE";
  if (allowed_require & REQUIRE_COMPARATOR_ASCII_NUMERIC) 
    capa[n++] = "COMPARATOR-I;ASCII-NUMERIC";
  if (allowed_require & REQUIRE_RELATIONAL) 
    capa[n++] = "RELATIONAL";
  if (allowed_require & REQUIRE_SUBADDRESS)
    capa[n++] = "SUBADDRESS";
  if (allowed_require & REQUIRE_COPY)
    capa[n++] = "COPY";
  if (allowed_require & REQUIRE_VIRUSTEST) 
    capa[n++] = "VIRUSTEST";
  if (allowed_require & REQUIRE_SPAMTEST) 
    capa[n++] = "SPAMTEST";
#ifdef WITH_SIEVE_REGEX
  if (allowed_require & REQUIRE_REGEX) 
    capa[n++] = "REGEX";
#endif
  capa[n++] = NULL;
  return capa;
}

